# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The rapidart module provides routines for artifact detection and region of
interest analysis.

These functions include:

  * ArtifactDetect: performs artifact detection on functional images

  * StimulusCorrelation: determines correlation between stimuli
    schedule and movement/intensity parameters
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import open, range, str, bytes

import os
from copy import deepcopy

from nibabel import load, funcs, Nifti1Image
import numpy as np
from scipy import signal
import scipy.io as sio

from ..utils import NUMPY_MMAP
from ..interfaces.base import (BaseInterface, traits, InputMultiPath,
                               OutputMultiPath, TraitedSpec, File,
                               BaseInterfaceInputSpec, isdefined)
from ..utils.filemanip import ensure_list, save_json, split_filename
from ..utils.misc import find_indices, normalize_mc_params
from .. import logging, config
iflogger = logging.getLogger('interface')


def _get_affine_matrix(params, source):
    """Return affine matrix given a set of translation and rotation parameters

    params : np.array (upto 12 long) in native package format
    source : the package that generated the parameters
             supports SPM, AFNI, FSFAST, FSL, NIPY
    """
    if source == 'NIPY':
        # nipy does not store typical euler angles, use nipy to convert
        from nipy.algorithms.registration import to_matrix44
        return to_matrix44(params)

    params = normalize_mc_params(params, source)
    # process for FSL, SPM, AFNI and FSFAST
    rotfunc = lambda x: np.array([[np.cos(x), np.sin(x)],
                                  [-np.sin(x), np.cos(x)]])
    q = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0])
    if len(params) < 12:
        params = np.hstack((params, q[len(params):]))
    params.shape = (len(params), )
    # Translation
    T = np.eye(4)
    T[0:3, -1] = params[0:3]
    # Rotation
    Rx = np.eye(4)
    Rx[1:3, 1:3] = rotfunc(params[3])
    Ry = np.eye(4)
    Ry[(0, 0, 2, 2), (0, 2, 0, 2)] = rotfunc(params[4]).ravel()
    Rz = np.eye(4)
    Rz[0:2, 0:2] = rotfunc(params[5])
    # Scaling
    S = np.eye(4)
    S[0:3, 0:3] = np.diag(params[6:9])
    # Shear
    Sh = np.eye(4)
    Sh[(0, 0, 1), (1, 2, 2)] = params[9:12]
    if source in ('AFNI', 'FSFAST'):
        return np.dot(T, np.dot(Ry, np.dot(Rx, np.dot(Rz, np.dot(S, Sh)))))
    return np.dot(T, np.dot(Rx, np.dot(Ry, np.dot(Rz, np.dot(S, Sh)))))


def _calc_norm(mc, use_differences, source, brain_pts=None):
    """Calculates the maximum overall displacement of the midpoints
    of the faces of a cube due to translation and rotation.

    Parameters
    ----------
    mc : motion parameter estimates
        [3 translation, 3 rotation (radians)]
    use_differences : boolean
    brain_pts : [4 x n_points] of coordinates

    Returns
    -------

    norm : at each time point
    displacement : euclidean distance (mm) of displacement at each coordinate

    """

    affines = [
        _get_affine_matrix(mc[i, :], source) for i in range(mc.shape[0])
    ]
    return _calc_norm_affine(affines, use_differences, brain_pts)


def _calc_norm_affine(affines, use_differences, brain_pts=None):
    """Calculates the maximum overall displacement of the midpoints
    of the faces of a cube due to translation and rotation.

    Parameters
    ----------
    affines : list of [4 x 4] affine matrices
    use_differences : boolean
    brain_pts : [4 x n_points] of coordinates

    Returns
    -------

    norm : at each time point
    displacement : euclidean distance (mm) of displacement at each coordinate

    """

    if brain_pts is None:
        respos = np.diag([70, 70, 75])
        resneg = np.diag([-70, -110, -45])
        all_pts = np.vstack((np.hstack((respos, resneg)), np.ones((1, 6))))
        displacement = None
    else:
        all_pts = brain_pts
    n_pts = all_pts.size - all_pts.shape[1]
    newpos = np.zeros((len(affines), n_pts))
    if brain_pts is not None:
        displacement = np.zeros((len(affines), int(n_pts / 3)))
    for i, affine in enumerate(affines):
        newpos[i, :] = np.dot(affine, all_pts)[0:3, :].ravel()
        if brain_pts is not None:
            displacement[i, :] = np.sqrt(
                np.sum(
                    np.power(
                        np.reshape(newpos[i, :],
                                   (3, all_pts.shape[1])) - all_pts[0:3, :],
                        2),
                    axis=0))
    # np.savez('displacement.npz', newpos=newpos, pts=all_pts)
    normdata = np.zeros(len(affines))
    if use_differences:
        newpos = np.concatenate(
            (np.zeros((1, n_pts)), np.diff(newpos, n=1, axis=0)), axis=0)
        for i in range(newpos.shape[0]):
            normdata[i] = \
                np.max(np.sqrt(np.sum(
                    np.reshape(np.power(np.abs(newpos[i, :]), 2),
                               (3, all_pts.shape[1])),
                    axis=0)))
    else:
        newpos = np.abs(signal.detrend(newpos, axis=0, type='constant'))
        normdata = np.sqrt(np.mean(np.power(newpos, 2), axis=1))
    return normdata, displacement


class ArtifactDetectInputSpec(BaseInterfaceInputSpec):
    realigned_files = InputMultiPath(
        File(exists=True),
        desc=("Names of realigned functional data "
              "files"),
        mandatory=True)
    realignment_parameters = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc=("Names of realignment "
              "parameters corresponding to "
              "the functional data files"))
    parameter_source = traits.Enum(
        "SPM",
        "FSL",
        "AFNI",
        "NiPy",
        "FSFAST",
        desc="Source of movement parameters",
        mandatory=True)
    use_differences = traits.ListBool(
        [True, False],
        minlen=2,
        maxlen=2,
        usedefault=True,
        desc=("Use differences between successive"
              " motion (first element) and "
              "intensity parameter (second "
              "element) estimates in order to "
              "determine outliers.  "
              "(default is [True, False])"))
    use_norm = traits.Bool(
        True,
        usedefault=True,
        requires=['norm_threshold'],
        desc=("Uses a composite of the motion parameters in "
              "order to determine outliers."))
    norm_threshold = traits.Float(
        xor=['rotation_threshold', 'translation_threshold'],
        mandatory=True,
        desc=("Threshold to use to detect motion-rela"
              "ted outliers when composite motion is "
              "being used"))
    rotation_threshold = traits.Float(
        mandatory=True,
        xor=['norm_threshold'],
        desc=("Threshold (in radians) to use to "
              "detect rotation-related outliers"))
    translation_threshold = traits.Float(
        mandatory=True,
        xor=['norm_threshold'],
        desc=("Threshold (in mm) to use to "
              "detect translation-related "
              "outliers"))
    zintensity_threshold = traits.Float(
        mandatory=True,
        desc=("Intensity Z-threshold use to "
              "detection images that deviate "
              "from the mean"))
    mask_type = traits.Enum(
        'spm_global',
        'file',
        'thresh',
        mandatory=True,
        desc=("Type of mask that should be used to mask the"
              " functional data. *spm_global* uses an "
              "spm_global like calculation to determine the"
              " brain mask. *file* specifies a brain mask "
              "file (should be an image file consisting of "
              "0s and 1s). *thresh* specifies a threshold "
              "to use. By default all voxels are used,"
              "unless one of these mask types are defined"))
    mask_file = File(
        exists=True, desc="Mask file to be used if mask_type is 'file'.")
    mask_threshold = traits.Float(
        desc=("Mask threshold to be used if mask_type"
              " is 'thresh'."))
    intersect_mask = traits.Bool(
        True, usedefault=True,
        desc=("Intersect the masks when computed from "
              "spm_global."))
    save_plot = traits.Bool(
        True, desc="save plots containing outliers", usedefault=True)
    plot_type = traits.Enum(
        'png',
        'svg',
        'eps',
        'pdf',
        desc="file type of the outlier plot",
        usedefault=True)
    bound_by_brainmask = traits.Bool(
        False,
        desc=("use the brain mask to "
              "determine bounding box"
              "for composite norm (works"
              "for SPM and Nipy - currently"
              "inaccurate for FSL, AFNI"),
        usedefault=True)
    global_threshold = traits.Float(
        8.0,
        desc=("use this threshold when mask "
              "type equal's spm_global"),
        usedefault=True)


class ArtifactDetectOutputSpec(TraitedSpec):
    outlier_files = OutputMultiPath(
        File(exists=True),
        desc=("One file for each functional run "
              "containing a list of 0-based indices"
              " corresponding to outlier volumes"))
    intensity_files = OutputMultiPath(
        File(exists=True),
        desc=("One file for each functional run "
              "containing the global intensity "
              "values determined from the "
              "brainmask"))
    norm_files = OutputMultiPath(
        File,
        desc=("One file for each functional run "
              "containing the composite norm"))
    statistic_files = OutputMultiPath(
        File(exists=True),
        desc=("One file for each functional run "
              "containing information about the "
              "different types of artifacts and "
              "if design info is provided then "
              "details of stimulus correlated "
              "motion and a listing or artifacts "
              "by event type."))
    plot_files = OutputMultiPath(
        File,
        desc=("One image file for each functional run "
              "containing the detected outliers"))
    mask_files = OutputMultiPath(
        File,
        desc=("One image file for each functional run "
              "containing the mask used for global "
              "signal calculation"))
    displacement_files = OutputMultiPath(
        File,
        desc=("One image file for each "
              "functional run containing the "
              "voxel displacement timeseries"))


class ArtifactDetect(BaseInterface):
    """Detects outliers in a functional imaging series

    Uses intensity and motion parameters to infer outliers. If `use_norm` is
    True, it computes the movement of the center of each face a cuboid centered
    around the head and returns the maximal movement across the centers. If you
    wish to use individual thresholds instead, import `Undefined` from
    `nipype.interfaces.base` and set `....inputs.use_norm = Undefined`


    Examples
    --------

    >>> ad = ArtifactDetect()
    >>> ad.inputs.realigned_files = 'functional.nii'
    >>> ad.inputs.realignment_parameters = 'functional.par'
    >>> ad.inputs.parameter_source = 'FSL'
    >>> ad.inputs.norm_threshold = 1
    >>> ad.inputs.use_differences = [True, False]
    >>> ad.inputs.zintensity_threshold = 3
    >>> ad.run()  # doctest: +SKIP
    """

    input_spec = ArtifactDetectInputSpec
    output_spec = ArtifactDetectOutputSpec

    def __init__(self, **inputs):
        super(ArtifactDetect, self).__init__(**inputs)

    def _get_output_filenames(self, motionfile, output_dir):
        """Generate output files based on motion filenames

        Parameters
        ----------

        motionfile: file/string
            Filename for motion parameter file
        output_dir: string
            output directory in which the files will be generated
        """
        if isinstance(motionfile, (str, bytes)):
            infile = motionfile
        elif isinstance(motionfile, list):
            infile = motionfile[0]
        else:
            raise Exception("Unknown type of file")
        _, filename, ext = split_filename(infile)
        artifactfile = os.path.join(output_dir, ''.join(('art.', filename,
                                                         '_outliers.txt')))
        intensityfile = os.path.join(output_dir, ''.join(('global_intensity.',
                                                          filename, '.txt')))
        statsfile = os.path.join(output_dir, ''.join(('stats.', filename,
                                                      '.txt')))
        normfile = os.path.join(output_dir, ''.join(('norm.', filename,
                                                     '.txt')))
        plotfile = os.path.join(output_dir, ''.join(('plot.', filename, '.',
                                                     self.inputs.plot_type)))
        displacementfile = os.path.join(output_dir, ''.join(('disp.', filename,
                                                             ext)))
        maskfile = os.path.join(output_dir, ''.join(('mask.', filename, ext)))
        return (artifactfile, intensityfile, statsfile, normfile, plotfile,
                displacementfile, maskfile)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['outlier_files'] = []
        outputs['intensity_files'] = []
        outputs['statistic_files'] = []
        outputs['mask_files'] = []
        if isdefined(self.inputs.use_norm) and self.inputs.use_norm:
            outputs['norm_files'] = []
            if self.inputs.bound_by_brainmask:
                outputs['displacement_files'] = []
        if isdefined(self.inputs.save_plot) and self.inputs.save_plot:
            outputs['plot_files'] = []
        for i, f in enumerate(ensure_list(self.inputs.realigned_files)):
            (outlierfile, intensityfile, statsfile, normfile, plotfile,
             displacementfile, maskfile) = \
                self._get_output_filenames(f, os.getcwd())
            outputs['outlier_files'].insert(i, outlierfile)
            outputs['intensity_files'].insert(i, intensityfile)
            outputs['statistic_files'].insert(i, statsfile)
            outputs['mask_files'].insert(i, maskfile)
            if isdefined(self.inputs.use_norm) and self.inputs.use_norm:
                outputs['norm_files'].insert(i, normfile)
                if self.inputs.bound_by_brainmask:
                    outputs['displacement_files'].insert(i, displacementfile)
            if isdefined(self.inputs.save_plot) and self.inputs.save_plot:
                outputs['plot_files'].insert(i, plotfile)
        return outputs

    def _plot_outliers_with_wave(self, wave, outliers, name):
        import matplotlib
        matplotlib.use(config.get("execution", "matplotlib_backend"))
        import matplotlib.pyplot as plt
        plt.plot(wave)
        plt.ylim([wave.min(), wave.max()])
        plt.xlim([0, len(wave) - 1])
        if len(outliers):
            plt.plot(
                np.tile(outliers[:, None], (1, 2)).T,
                np.tile([wave.min(), wave.max()], (len(outliers), 1)).T, 'r')
        plt.xlabel('Scans - 0-based')
        plt.ylabel(name)

    def _detect_outliers_core(self, imgfile, motionfile, runidx, cwd=None):
        """
        Core routine for detecting outliers
        """
        if not cwd:
            cwd = os.getcwd()

        # read in functional image
        if isinstance(imgfile, (str, bytes)):
            nim = load(imgfile, mmap=NUMPY_MMAP)
        elif isinstance(imgfile, list):
            if len(imgfile) == 1:
                nim = load(imgfile[0], mmap=NUMPY_MMAP)
            else:
                images = [load(f, mmap=NUMPY_MMAP) for f in imgfile]
                nim = funcs.concat_images(images)

        # compute global intensity signal
        (x, y, z, timepoints) = nim.shape

        data = nim.get_data()
        affine = nim.affine
        g = np.zeros((timepoints, 1))
        masktype = self.inputs.mask_type
        if masktype == 'spm_global':  # spm_global like calculation
            iflogger.debug('art: using spm global')
            intersect_mask = self.inputs.intersect_mask
            if intersect_mask:
                mask = np.ones((x, y, z), dtype=bool)
                for t0 in range(timepoints):
                    vol = data[:, :, :, t0]
                    # Use an SPM like approach
                    mask_tmp = vol > \
                        (np.nanmean(vol) / self.inputs.global_threshold)
                    mask = mask * mask_tmp
                for t0 in range(timepoints):
                    vol = data[:, :, :, t0]
                    g[t0] = np.nanmean(vol[mask])
                if len(find_indices(mask)) < (np.prod((x, y, z)) / 10):
                    intersect_mask = False
                    g = np.zeros((timepoints, 1))
            if not intersect_mask:
                iflogger.info('not intersect_mask is True')
                mask = np.zeros((x, y, z, timepoints))
                for t0 in range(timepoints):
                    vol = data[:, :, :, t0]
                    mask_tmp = vol > \
                        (np.nanmean(vol) / self.inputs.global_threshold)
                    mask[:, :, :, t0] = mask_tmp
                    g[t0] = np.nansum(vol * mask_tmp) / np.nansum(mask_tmp)
        elif masktype == 'file':  # uses a mask image to determine intensity
            maskimg = load(self.inputs.mask_file, mmap=NUMPY_MMAP)
            mask = maskimg.get_data()
            affine = maskimg.affine
            mask = mask > 0.5
            for t0 in range(timepoints):
                vol = data[:, :, :, t0]
                g[t0] = np.nanmean(vol[mask])
        elif masktype == 'thresh':  # uses a fixed signal threshold
            for t0 in range(timepoints):
                vol = data[:, :, :, t0]
                mask = vol > self.inputs.mask_threshold
                g[t0] = np.nanmean(vol[mask])
        else:
            mask = np.ones((x, y, z))
            g = np.nanmean(data[mask > 0, :], 1)

        # compute normalized intensity values
        gz = signal.detrend(g, axis=0)  # detrend the signal
        if self.inputs.use_differences[1]:
            gz = np.concatenate(
                (np.zeros((1, 1)), np.diff(gz, n=1, axis=0)), axis=0)
        gz = (gz - np.mean(gz)) / np.std(gz)  # normalize the detrended signal
        iidx = find_indices(abs(gz) > self.inputs.zintensity_threshold)

        # read in motion parameters
        mc_in = np.loadtxt(motionfile)
        mc = deepcopy(mc_in)

        (artifactfile, intensityfile, statsfile, normfile, plotfile,
         displacementfile, maskfile) = self._get_output_filenames(
             imgfile, cwd)
        mask_img = Nifti1Image(mask.astype(np.uint8), affine)
        mask_img.to_filename(maskfile)

        if self.inputs.use_norm:
            brain_pts = None
            if self.inputs.bound_by_brainmask:
                voxel_coords = np.nonzero(mask)
                coords = np.vstack((voxel_coords[0],
                                    np.vstack((voxel_coords[1],
                                               voxel_coords[2])))).T
                brain_pts = np.dot(affine,
                                   np.hstack((coords,
                                              np.ones((coords.shape[0],
                                                       1)))).T)
            # calculate the norm of the motion parameters
            normval, displacement = _calc_norm(
                mc,
                self.inputs.use_differences[0],
                self.inputs.parameter_source,
                brain_pts=brain_pts)
            tidx = find_indices(normval > self.inputs.norm_threshold)
            ridx = find_indices(normval < 0)
            if displacement is not None:
                dmap = np.zeros((x, y, z, timepoints), dtype=np.float)
                for i in range(timepoints):
                    dmap[voxel_coords[0], voxel_coords[1], voxel_coords[2],
                         i] = displacement[i, :]
                dimg = Nifti1Image(dmap, affine)
                dimg.to_filename(displacementfile)
        else:
            if self.inputs.use_differences[0]:
                mc = np.concatenate(
                    (np.zeros((1, 6)), np.diff(mc_in, n=1, axis=0)), axis=0)
            traval = mc[:, 0:3]  # translation parameters (mm)
            rotval = mc[:, 3:6]  # rotation parameters (rad)
            tidx = find_indices(
                np.sum(abs(traval) > self.inputs.translation_threshold, 1) > 0)
            ridx = find_indices(
                np.sum(abs(rotval) > self.inputs.rotation_threshold, 1) > 0)

        outliers = np.unique(np.union1d(iidx, np.union1d(tidx, ridx)))

        # write output to outputfile
        np.savetxt(artifactfile, outliers, fmt=b'%d', delimiter=' ')
        np.savetxt(intensityfile, g, fmt=b'%.2f', delimiter=' ')
        if self.inputs.use_norm:
            np.savetxt(normfile, normval, fmt=b'%.4f', delimiter=' ')

        if isdefined(self.inputs.save_plot) and self.inputs.save_plot:
            import matplotlib
            matplotlib.use(config.get("execution", "matplotlib_backend"))
            import matplotlib.pyplot as plt
            fig = plt.figure()
            if isdefined(self.inputs.use_norm) and self.inputs.use_norm:
                plt.subplot(211)
            else:
                plt.subplot(311)
            self._plot_outliers_with_wave(gz, iidx, 'Intensity')
            if isdefined(self.inputs.use_norm) and self.inputs.use_norm:
                plt.subplot(212)
                self._plot_outliers_with_wave(normval, np.union1d(tidx, ridx),
                                              'Norm (mm)')
            else:
                diff = ''
                if self.inputs.use_differences[0]:
                    diff = 'diff'
                plt.subplot(312)
                self._plot_outliers_with_wave(traval, tidx,
                                              'Translation (mm)' + diff)
                plt.subplot(313)
                self._plot_outliers_with_wave(rotval, ridx,
                                              'Rotation (rad)' + diff)
            plt.savefig(plotfile)
            plt.close(fig)

        motion_outliers = np.union1d(tidx, ridx)
        stats = [
            {
                'motion_file': motionfile,
                'functional_file': imgfile
            },
            {
                'common_outliers': len(np.intersect1d(iidx, motion_outliers)),
                'intensity_outliers': len(np.setdiff1d(iidx, motion_outliers)),
                'motion_outliers': len(np.setdiff1d(motion_outliers, iidx)),
            },
            {
                'motion': [
                    {
                        'using differences': self.inputs.use_differences[0]
                    },
                    {
                        'mean': np.mean(mc_in, axis=0).tolist(),
                        'min': np.min(mc_in, axis=0).tolist(),
                        'max': np.max(mc_in, axis=0).tolist(),
                        'std': np.std(mc_in, axis=0).tolist()
                    },
                ]
            },
            {
                'intensity': [
                    {
                        'using differences': self.inputs.use_differences[1]
                    },
                    {
                        'mean': np.mean(gz, axis=0).tolist(),
                        'min': np.min(gz, axis=0).tolist(),
                        'max': np.max(gz, axis=0).tolist(),
                        'std': np.std(gz, axis=0).tolist()
                    },
                ]
            },
        ]
        if self.inputs.use_norm:
            stats.insert(
                3, {
                    'motion_norm': {
                        'mean': np.mean(normval, axis=0).tolist(),
                        'min': np.min(normval, axis=0).tolist(),
                        'max': np.max(normval, axis=0).tolist(),
                        'std': np.std(normval, axis=0).tolist(),
                    }
                })
        save_json(statsfile, stats)

    def _run_interface(self, runtime):
        """Execute this module.
        """
        funcfilelist = ensure_list(self.inputs.realigned_files)
        motparamlist = ensure_list(self.inputs.realignment_parameters)
        for i, imgf in enumerate(funcfilelist):
            self._detect_outliers_core(
                imgf, motparamlist[i], i, cwd=os.getcwd())
        return runtime


class StimCorrInputSpec(BaseInterfaceInputSpec):
    realignment_parameters = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc=("Names of realignment "
              "parameters corresponding to "
              "the functional data files"))
    intensity_values = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc=("Name of file containing intensity "
              "values"))
    spm_mat_file = File(
        exists=True,
        mandatory=True,
        desc="SPM mat file (use pre-estimate SPM.mat file)")
    concatenated_design = traits.Bool(
        mandatory=True,
        desc=("state if the design matrix "
              "contains concatenated sessions"))


class StimCorrOutputSpec(TraitedSpec):
    stimcorr_files = OutputMultiPath(
        File(exists=True),
        desc=("List of files containing "
              "correlation values"))


class StimulusCorrelation(BaseInterface):
    """Determines if stimuli are correlated with motion or intensity
    parameters.

    Currently this class supports an SPM generated design matrix and requires
    intensity parameters. This implies that one must run
    :ref:`ArtifactDetect <nipype.algorithms.rapidart.ArtifactDetect>`
    and :ref:`Level1Design <nipype.interfaces.spm.model.Level1Design>` prior to
    running this or provide an SPM.mat file and intensity parameters through
    some other means.

    Examples
    --------

    >>> sc = StimulusCorrelation()
    >>> sc.inputs.realignment_parameters = 'functional.par'
    >>> sc.inputs.intensity_values = 'functional.rms'
    >>> sc.inputs.spm_mat_file = 'SPM.mat'
    >>> sc.inputs.concatenated_design = False
    >>> sc.run() # doctest: +SKIP

    """

    input_spec = StimCorrInputSpec
    output_spec = StimCorrOutputSpec

    def _get_output_filenames(self, motionfile, output_dir):
        """Generate output files based on motion filenames

        Parameters
        ----------
        motionfile: file/string
            Filename for motion parameter file
        output_dir: string
            output directory in which the files will be generated
        """
        (_, filename) = os.path.split(motionfile)
        (filename, _) = os.path.splitext(filename)
        corrfile = os.path.join(output_dir, ''.join(('qa.', filename,
                                                     '_stimcorr.txt')))
        return corrfile

    def _stimcorr_core(self, motionfile, intensityfile, designmatrix,
                       cwd=None):
        """
        Core routine for determining stimulus correlation

        """
        if not cwd:
            cwd = os.getcwd()
        # read in motion parameters
        mc_in = np.loadtxt(motionfile)
        g_in = np.loadtxt(intensityfile)
        g_in.shape = g_in.shape[0], 1
        dcol = designmatrix.shape[1]
        mccol = mc_in.shape[1]
        concat_matrix = np.hstack((np.hstack((designmatrix, mc_in)), g_in))
        cm = np.corrcoef(concat_matrix, rowvar=0)
        corrfile = self._get_output_filenames(motionfile, cwd)
        # write output to outputfile
        file = open(corrfile, 'w')
        file.write("Stats for:\n")
        file.write("Stimulus correlated motion:\n%s\n" % motionfile)
        for i in range(dcol):
            file.write("SCM.%d:" % i)
            for v in cm[i, dcol + np.arange(mccol)]:
                file.write(" %.2f" % v)
            file.write('\n')
        file.write("Stimulus correlated intensity:\n%s\n" % intensityfile)
        for i in range(dcol):
            file.write("SCI.%d: %.2f\n" % (i, cm[i, -1]))
        file.close()

    def _get_spm_submatrix(self, spmmat, sessidx, rows=None):
        """
        Parameters
        ----------
        spmmat: scipy matlab object
            full SPM.mat file loaded into a scipy object
        sessidx: int
            index to session that needs to be extracted.
        """
        designmatrix = spmmat['SPM'][0][0].xX[0][0].X
        U = spmmat['SPM'][0][0].Sess[0][sessidx].U[0]
        if rows is None:
            rows = spmmat['SPM'][0][0].Sess[0][sessidx].row[0] - 1
        cols = (spmmat['SPM'][0][0].Sess[0][sessidx].col[0][list(
            range(len(U)))] - 1)
        outmatrix = designmatrix.take(
            rows.tolist(), axis=0).take(
                cols.tolist(), axis=1)
        return outmatrix

    def _run_interface(self, runtime):
        """Execute this module.
        """
        motparamlist = self.inputs.realignment_parameters
        intensityfiles = self.inputs.intensity_values
        spmmat = sio.loadmat(self.inputs.spm_mat_file, struct_as_record=False)
        nrows = []
        for i in range(len(motparamlist)):
            sessidx = i
            rows = None
            if self.inputs.concatenated_design:
                sessidx = 0
                mc_in = np.loadtxt(motparamlist[i])
                rows = np.sum(nrows) + np.arange(mc_in.shape[0])
                nrows.append(mc_in.shape[0])
            matrix = self._get_spm_submatrix(spmmat, sessidx, rows)
            self._stimcorr_core(motparamlist[i], intensityfiles[i], matrix,
                                os.getcwd())
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        files = []
        for i, f in enumerate(self.inputs.realignment_parameters):
            files.insert(i, self._get_output_filenames(f, os.getcwd()))
        if files:
            outputs['stimcorr_files'] = files
        return outputs



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from ....utils import NUMPY_MMAP
from ....pipeline import engine as pe
from ....interfaces.utility import Function, IdentityInterface
from ....interfaces.freesurfer import *
from .utils import copy_file


def checkT1s(T1_files, cw256=False):
    """Verifying size of inputs and setting workflow parameters"""
    import sys
    import nibabel as nb
    from nipype.utils.filemanip import ensure_list

    T1_files = ensure_list(T1_files)
    if len(T1_files) == 0:
        print("ERROR: No T1's Given")
        sys.exit(-1)

    shape = nb.load(T1_files[0]).shape
    for t1 in T1_files[1:]:
        if nb.load(t1, mmap=NUMPY_MMAP).shape != shape:
            print("ERROR: T1s not the same size. Cannot process {0} and {1} "
                  "together".format(T1_files[0], t1))
            sys.exit(-1)

    origvol_names = ["{0:03d}.mgz".format(i + 1) for i in range(len(T1_files))]

    # check if cw256 is set to crop the images if size is larger than 256
    if not cw256 and any(dim > 256 for dim in shape):
        print("Setting MRI Convert to crop images to 256 FOV")
        cw256 = True

    resample_type = 'cubic' if len(T1_files) > 1 else 'interpolate'
    return T1_files, cw256, resample_type, origvol_names


def create_AutoRecon1(name="AutoRecon1",
                      longitudinal=False,
                      distance=None,
                      custom_atlas=None,
                      plugin_args=None,
                      shrink=None,
                      stop=None,
                      fsvernum=5.3):
    """Creates the AutoRecon1 workflow in nipype.

    Inputs::
           inputspec.T1_files : T1 files (mandatory)
           inputspec.T2_file : T2 file (optional)
           inputspec.FLAIR_file : FLAIR file (optional)
           inputspec.cw256 : Conform inputs to 256 FOV (optional)
           inputspec.num_threads: Number of threads to use with EM Register (default=1)
    Outpus::

    """
    ar1_wf = pe.Workflow(name=name)
    inputspec = pe.Node(
        interface=IdentityInterface(fields=[
            'T1_files', 'T2_file', 'FLAIR_file', 'cw256', 'num_threads',
            'reg_template_withskull', 'awk_file'
        ]),
        run_without_submitting=True,
        name='inputspec')

    if not longitudinal:
        # single session processing
        verify_inputs = pe.Node(
            Function(["T1_files", "cw256"],
                     ["T1_files", "cw256", "resample_type", "origvol_names"],
                     checkT1s),
            name="Check_T1s")
        ar1_wf.connect([(inputspec, verify_inputs, [('T1_files', 'T1_files'),
                                                    ('cw256', 'cw256')])])

        # T1 image preparation
        # For all T1's mri_convert ${InputVol} ${out_file}
        T1_image_preparation = pe.MapNode(
            MRIConvert(), iterfield=['in_file', 'out_file'], name="T1_prep")

        ar1_wf.connect([
            (verify_inputs, T1_image_preparation,
             [('T1_files', 'in_file'), ('origvol_names', 'out_file')]),
        ])

        def convert_modalities(in_file=None, out_file=None):
            """Returns an undefined output if the in_file is not defined"""
            from nipype.interfaces.freesurfer import MRIConvert
            import os
            if in_file:
                convert = MRIConvert()
                convert.inputs.in_file = in_file
                convert.inputs.out_file = out_file
                convert.inputs.no_scale = True
                out = convert.run()
                out_file = os.path.abspath(out.outputs.out_file)
            return out_file

        T2_convert = pe.Node(
            Function(['in_file', 'out_file'], ['out_file'],
                     convert_modalities),
            name="T2_Convert")
        T2_convert.inputs.out_file = 'T2raw.mgz'
        ar1_wf.connect([(inputspec, T2_convert, [('T2_file', 'in_file')])])

        FLAIR_convert = pe.Node(
            Function(['in_file', 'out_file'], ['out_file'],
                     convert_modalities),
            name="FLAIR_Convert")
        FLAIR_convert.inputs.out_file = 'FLAIRraw.mgz'
        ar1_wf.connect([(inputspec, FLAIR_convert, [('FLAIR_file',
                                                     'in_file')])])
    else:
        # longitudinal inputs
        inputspec = pe.Node(
            interface=IdentityInterface(fields=[
                'T1_files', 'iscales', 'ltas', 'subj_to_template_lta',
                'template_talairach_xfm', 'template_brainmask'
            ]),
            run_without_submitting=True,
            name='inputspec')

        def output_names(T1_files):
            """Create file names that are dependent on the number of T1 inputs"""
            iscale_names = list()
            lta_names = list()
            for i, t1 in enumerate(T1_files):
                # assign an input number
                file_num = str(i + 1)
                while len(file_num) < 3:
                    file_num = '0' + file_num
                iscale_names.append("{0}-iscale.txt".format(file_num))
                lta_names.append("{0}.lta".format(file_num))
            return iscale_names, lta_names

        filenames = pe.Node(
            Function(['T1_files'], ['iscale_names', 'lta_names'],
                     output_names),
            name="Longitudinal_Filenames")
        ar1_wf.connect([(inputspec, filenames, [('T1_files', 'T1_files')])])

        copy_ltas = pe.MapNode(
            Function(['in_file', 'out_file'], ['out_file'], copy_file),
            iterfield=['in_file', 'out_file'],
            name='Copy_ltas')
        ar1_wf.connect([(inputspec, copy_ltas, [('ltas', 'in_file')]),
                        (filenames, copy_ltas, [('lta_names', 'out_file')])])

        copy_iscales = pe.MapNode(
            Function(['in_file', 'out_file'], ['out_file'], copy_file),
            iterfield=['in_file', 'out_file'],
            name='Copy_iscales')
        ar1_wf.connect([(inputspec, copy_iscales, [('iscales', 'in_file')]),
                        (filenames, copy_iscales, [('iscale_names',
                                                    'out_file')])])

        concatenate_lta = pe.MapNode(
            ConcatenateLTA(), iterfield=['in_file'], name="Concatenate_ltas")
        ar1_wf.connect([(copy_ltas, concatenate_lta, [('out_file',
                                                       'in_file')]),
                        (inputspec, concatenate_lta, [('subj_to_template_lta',
                                                       'subj_to_base')])])

    # Motion Correction
    """
    When there are multiple source volumes, this step will correct for small
    motions between them and then average them together.  The output of the
    motion corrected average is mri/rawavg.mgz which is then conformed to
    255 cubed char images (1mm isotropic voxels) in mri/orig.mgz.
    """

    def createTemplate(in_files, out_file):
        import os
        import shutil
        if len(in_files) == 1:
            # if only 1 T1 scan given, no need to run RobustTemplate
            print(
                "WARNING: only one run found. This is OK, but motion correction "
                +
                "cannot be performed on one run, so I'll copy the run to rawavg "
                + "and continue.")
            shutil.copyfile(in_files[0], out_file)
            intensity_scales = None
            transforms = None
        else:
            from nipype.interfaces.freesurfer import RobustTemplate
            # if multiple T1 scans are given run RobustTemplate
            intensity_scales = [
                os.path.basename(f.replace('.mgz', '-iscale.txt'))
                for f in in_files
            ]
            transforms = [
                os.path.basename(f.replace('.mgz', '.lta')) for f in in_files
            ]
            robtemp = RobustTemplate()
            robtemp.inputs.in_files = in_files
            robtemp.inputs.average_metric = 'median'
            robtemp.inputs.out_file = out_file
            robtemp.inputs.no_iteration = True
            robtemp.inputs.fixed_timepoint = True
            robtemp.inputs.auto_detect_sensitivity = True
            robtemp.inputs.initial_timepoint = 1
            robtemp.inputs.scaled_intensity_outputs = intensity_scales
            robtemp.inputs.transform_outputs = transforms
            robtemp.inputs.subsample_threshold = 200
            robtemp.inputs.intensity_scaling = True
            robtemp_result = robtemp.run()
            # collect the outputs from RobustTemplate
            out_file = robtemp_result.outputs.out_file
            intensity_scales = [
                os.path.abspath(f)
                for f in robtemp_result.outputs.scaled_intensity_outputs
            ]
            transforms = [
                os.path.abspath(f)
                for f in robtemp_result.outputs.transform_outputs
            ]
        out_file = os.path.abspath(out_file)
        return out_file, intensity_scales, transforms

    if not longitudinal:
        create_template = pe.Node(
            Function(['in_files', 'out_file'],
                     ['out_file', 'intensity_scales', 'transforms'],
                     createTemplate),
            name="Robust_Template")
        create_template.inputs.out_file = 'rawavg.mgz'
        ar1_wf.connect([(T1_image_preparation, create_template,
                         [('out_file', 'in_files')])])
    else:
        create_template = pe.Node(RobustTemplate(), name="Robust_Template")
        create_template.inputs.average_metric = 'median'
        create_template.inputs.out_file = 'rawavg.mgz'
        create_template.inputs.no_iteration = True
        ar1_wf.connect([(concatenate_lta, create_template,
                         [('out_file', 'initial_transforms')]),
                        (inputSpec, create_template, [('in_t1s', 'in_files')]),
                        (copy_iscales, create_template,
                         [('out_file', 'in_intensity_scales')])])

    # mri_convert
    conform_template = pe.Node(MRIConvert(), name='Conform_Template')
    conform_template.inputs.out_file = 'orig.mgz'
    if not longitudinal:
        conform_template.inputs.conform = True
        ar1_wf.connect([(verify_inputs, conform_template,
                         [('cw256', 'cw256'), ('resample_type',
                                               'resample_type')])])
    else:
        conform_template.inputs.out_datatype = 'uchar'

    ar1_wf.connect([(create_template, conform_template, [('out_file',
                                                          'in_file')])])

    # Talairach
    """
    This computes the affine transform from the orig volume to the MNI305 atlas using Avi Snyders 4dfp
    suite of image registration tools, through a FreeSurfer script called talairach_avi.
    Several of the downstream programs use talairach coordinates as seed points.
    """

    bias_correction = pe.Node(MNIBiasCorrection(), name="Bias_correction")
    bias_correction.inputs.iterations = 1
    bias_correction.inputs.protocol_iterations = 1000
    bias_correction.inputs.distance = distance
    if stop:
        bias_correction.inputs.stop = stop
    if shrink:
        bias_correction.inputs.shrink = shrink
    bias_correction.inputs.no_rescale = True
    bias_correction.inputs.out_file = 'orig_nu.mgz'

    ar1_wf.connect([
        (conform_template, bias_correction, [('out_file', 'in_file')]),
    ])

    if not longitudinal:
        # single session processing
        talairach_avi = pe.Node(TalairachAVI(), name="Compute_Transform")
        if custom_atlas is not None:
            # allows to specify a custom atlas
            talairach_avi.inputs.atlas = custom_atlas
        talairach_avi.inputs.out_file = 'talairach.auto.xfm'
        ar1_wf.connect([(bias_correction, talairach_avi, [('out_file',
                                                           'in_file')])])
    else:
        # longitudinal processing
        # Just copy the template xfm
        talairach_avi = pe.Node(
            Function(['in_file', 'out_file'], ['out_file'], copy_file),
            name='Copy_Template_Transform')
        talairach_avi.inputs.out_file = 'talairach.auto.xfm'

        ar1_wf.connect([(inputspec, talairach_avi, [('template_talairach_xfm',
                                                     'in_file')])])

    copy_transform = pe.Node(
        Function(['in_file', 'out_file'], ['out_file'], copy_file),
        name='Copy_Transform')
    copy_transform.inputs.out_file = 'talairach.xfm'

    ar1_wf.connect([(talairach_avi, copy_transform, [('out_file',
                                                      'in_file')])])

    # In recon-all the talairach.xfm is added to orig.mgz, even though
    # it does not exist yet. This is a compromise to keep from
    # having to change the time stamp of the orig volume after talairaching.
    # Here we are going to add xfm to the header after the xfm has been created.
    # This may mess up the timestamp.

    add_xform_to_orig = pe.Node(
        AddXFormToHeader(), name="Add_Transform_to_Orig")
    add_xform_to_orig.inputs.copy_name = True
    add_xform_to_orig.inputs.out_file = conform_template.inputs.out_file

    ar1_wf.connect(
        [(conform_template, add_xform_to_orig, [('out_file', 'in_file')]),
         (copy_transform, add_xform_to_orig, [('out_file', 'transform')])])

    # This node adds the transform to the orig_nu.mgz file. This step does not
    # exist in the recon-all workflow, because that workflow adds the talairach
    # to the orig.mgz file header before the talairach actually exists.
    add_xform_to_orig_nu = pe.Node(
        AddXFormToHeader(), name="Add_Transform_to_Orig_Nu")
    add_xform_to_orig_nu.inputs.copy_name = True
    add_xform_to_orig_nu.inputs.out_file = bias_correction.inputs.out_file

    ar1_wf.connect(
        [(bias_correction, add_xform_to_orig_nu, [('out_file', 'in_file')]),
         (copy_transform, add_xform_to_orig_nu, [('out_file', 'transform')])])

    # check the alignment of the talairach
    # TODO: Figure out how to read output from this node.
    check_alignment = pe.Node(
        CheckTalairachAlignment(), name="Check_Talairach_Alignment")
    check_alignment.inputs.threshold = 0.005
    ar1_wf.connect([
        (copy_transform, check_alignment, [('out_file', 'in_file')]),
    ])

    if not longitudinal:

        def awkfile(in_file, log_file):
            """
            This method uses 'awk' which must be installed prior to running the workflow and is not a
            part of nipype or freesurfer.
            """
            import subprocess
            import os
            command = ['awk', '-f', in_file, log_file]
            print(''.join(command))
            subprocess.call(command)
            log_file = os.path.abspath(log_file)
            return log_file

        awk_logfile = pe.Node(
            Function(['in_file', 'log_file'], ['log_file'], awkfile),
            name='Awk')

        ar1_wf.connect([(talairach_avi, awk_logfile, [('out_log',
                                                       'log_file')]),
                        (inputspec, awk_logfile, [('awk_file', 'in_file')])])

        # TODO datasink the output from TalirachQC...not sure how to do this
        tal_qc = pe.Node(TalairachQC(), name="Detect_Aligment_Failures")
        ar1_wf.connect([(awk_logfile, tal_qc, [('log_file', 'log_file')])])

    if fsvernum < 6:
        # intensity correction is performed before normalization
        intensity_correction = pe.Node(
            MNIBiasCorrection(), name="Intensity_Correction")
        intensity_correction.inputs.out_file = 'nu.mgz'
        intensity_correction.inputs.iterations = 2
        ar1_wf.connect([(add_xform_to_orig, intensity_correction,
                         [('out_file', 'in_file')]),
                        (copy_transform, intensity_correction,
                         [('out_file', 'transform')])])

        add_to_header_nu = pe.Node(AddXFormToHeader(), name="Add_XForm_to_NU")
        add_to_header_nu.inputs.copy_name = True
        add_to_header_nu.inputs.out_file = 'nu.mgz'
        ar1_wf.connect([(intensity_correction, add_to_header_nu, [
            ('out_file', 'in_file'),
        ]), (copy_transform, add_to_header_nu, [('out_file', 'transform')])])

    # Intensity Normalization
    # Performs intensity normalization of the orig volume and places the result in mri/T1.mgz.
    # Attempts to correct for fluctuations in intensity that would otherwise make intensity-based
    # segmentation much more difficult. Intensities for all voxels are scaled so that the mean
    # intensity of the white matter is 110.

    mri_normalize = pe.Node(Normalize(), name="Normalize_T1")
    mri_normalize.inputs.gradient = 1
    mri_normalize.inputs.out_file = 'T1.mgz'

    if fsvernum < 6:
        ar1_wf.connect([(add_to_header_nu, mri_normalize, [('out_file',
                                                            'in_file')])])
    else:
        ar1_wf.connect([(add_xform_to_orig_nu, mri_normalize, [('out_file',
                                                                'in_file')])])

    ar1_wf.connect([(copy_transform, mri_normalize, [('out_file',
                                                      'transform')])])

    # Skull Strip
    """
    Removes the skull from mri/T1.mgz and stores the result in
    mri/brainmask.auto.mgz and mri/brainmask.mgz. Runs the mri_watershed program.
    """
    if not longitudinal:
        mri_em_register = pe.Node(EMRegister(), name="EM_Register")
        mri_em_register.inputs.out_file = 'talairach_with_skull.lta'
        mri_em_register.inputs.skull = True
        if plugin_args:
            mri_em_register.plugin_args = plugin_args

        if fsvernum < 6:
            ar1_wf.connect(add_to_header_nu, 'out_file', mri_em_register,
                           'in_file')
        else:
            ar1_wf.connect(add_xform_to_orig_nu, 'out_file', mri_em_register,
                           'in_file')

        ar1_wf.connect([(inputspec, mri_em_register,
                         [('num_threads', 'num_threads'),
                          ('reg_template_withskull', 'template')])])

        brainmask = pe.Node(
            WatershedSkullStrip(), name='Watershed_Skull_Strip')
        brainmask.inputs.t1 = True
        brainmask.inputs.out_file = 'brainmask.auto.mgz'
        ar1_wf.connect([(mri_normalize, brainmask, [('out_file', 'in_file')]),
                        (mri_em_register, brainmask, [('out_file',
                                                       'transform')]),
                        (inputspec, brainmask, [('reg_template_withskull',
                                                 'brain_atlas')])])
    else:
        copy_template_brainmask = pe.Node(
            Function(['in_file', 'out_file'], ['out_file'], copy_file),
            name='Copy_Template_Brainmask')
        copy_template_brainmask.inputs.out_file = 'brainmask_{0}.mgz'.format(
            config['long_template'])

        ar1_wf.connect([(inputspec, copy_template_brainmask,
                         [('template_brainmask', 'in_file')])])

        mask1 = pe.Node(ApplyMask(), name="ApplyMask1")
        mask1.inputs.keep_mask_deletion_edits = True
        mask1.inputs.out_file = 'brainmask.auto.mgz'

        ar1_wf.connect([(mri_normalize, mask1, [('out_file', 'in_file')]),
                        (copy_template_brainmask, mask1, [('out_file',
                                                           'mask_file')])])

        brainmask = pe.Node(ApplyMask(), name="ApplyMask2")
        brainmask.inputs.keep_mask_deletion_edits = True
        brainmask.inputs.transfer = 255
        brainmask.inputs.out_file = mask1.inputs.out_file

        ar1_wf.connect([(mask1, brainmask, [('out_file', 'in_file')]),
                        (copy_template_brainmask, brainmask, [('out_file',
                                                               'mask_file')])])

    copy_brainmask = pe.Node(
        Function(['in_file', 'out_file'], ['out_file'], copy_file),
        name='Copy_Brainmask')
    copy_brainmask.inputs.out_file = 'brainmask.mgz'

    ar1_wf.connect([(brainmask, copy_brainmask, [('out_file', 'in_file')])])

    outputs = [
        'origvols', 't2_raw', 'flair', 'rawavg', 'orig_nu', 'orig',
        'talairach_auto', 'talairach', 't1', 'talskull', 'brainmask_auto',
        'brainmask', 'braintemplate'
    ]

    if fsvernum < 6:
        outputspec = pe.Node(
            IdentityInterface(fields=outputs + ['nu']), name="outputspec")
        ar1_wf.connect([(add_to_header_nu, outputspec, [('out_file', 'nu')])])
    else:
        outputspec = pe.Node(
            IdentityInterface(fields=outputs), name="outputspec")

    ar1_wf.connect([
        (T1_image_preparation, outputspec, [('out_file', 'origvols')]),
        (T2_convert, outputspec, [('out_file', 't2_raw')]),
        (FLAIR_convert, outputspec, [('out_file', 'flair')]),
        (create_template, outputspec, [('out_file', 'rawavg')]),
        (add_xform_to_orig, outputspec, [('out_file', 'orig')]),
        (add_xform_to_orig_nu, outputspec, [('out_file', 'orig_nu')]),
        (talairach_avi, outputspec, [('out_file', 'talairach_auto')]),
        (copy_transform, outputspec, [('out_file', 'talairach')]),
        (mri_normalize, outputspec, [('out_file', 't1')]),
        (brainmask, outputspec, [('out_file', 'brainmask_auto')]),
        (copy_brainmask, outputspec, [('out_file', 'brainmask')]),
    ])

    if not longitudinal:
        ar1_wf.connect([
            (mri_em_register, outputspec, [('out_file', 'talskull')]),
        ])
    else:
        ar1_wf.connect([
            (copy_template_brainmask, outputspec, [('out_file',
                                                    'braintemplate')]),
        ])

    return ar1_wf, outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..cmtk import ROIGen


def test_ROIGen_inputs():
    input_map = dict(
        LUT_file=dict(xor=['use_freesurfer_LUT'], ),
        aparc_aseg_file=dict(mandatory=True, ),
        freesurfer_dir=dict(requires=['use_freesurfer_LUT'], ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        out_dict_file=dict(genfile=True, ),
        out_roi_file=dict(genfile=True, ),
        use_freesurfer_LUT=dict(xor=['LUT_file'], ),
    )
    inputs = ROIGen.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ROIGen_outputs():
    output_map = dict(
        dict_file=dict(),
        roi_file=dict(),
    )
    outputs = ROIGen.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..developer import JistIntensityMp2rageMasking


def test_JistIntensityMp2rageMasking_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        inBackground=dict(argstr='--inBackground %s', ),
        inMasking=dict(argstr='--inMasking %s', ),
        inQuantitative=dict(argstr='--inQuantitative %s', ),
        inSecond=dict(argstr='--inSecond %s', ),
        inSkip=dict(argstr='--inSkip %s', ),
        inT1weighted=dict(argstr='--inT1weighted %s', ),
        null=dict(argstr='--null %s', ),
        outMasked=dict(
            argstr='--outMasked_T1_Map %s',
            hash_files=False,
        ),
        outMasked2=dict(
            argstr='--outMasked_T1weighted %s',
            hash_files=False,
        ),
        outSignal=dict(
            argstr='--outSignal_Proba %s',
            hash_files=False,
        ),
        outSignal2=dict(
            argstr='--outSignal_Mask %s',
            hash_files=False,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        xDefaultMem=dict(argstr='-xDefaultMem %d', ),
        xMaxProcess=dict(
            argstr='-xMaxProcess %d',
            usedefault=True,
        ),
        xPrefExt=dict(argstr='--xPrefExt %s', ),
    )
    inputs = JistIntensityMp2rageMasking.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_JistIntensityMp2rageMasking_outputs():
    output_map = dict(
        outMasked=dict(),
        outMasked2=dict(),
        outSignal=dict(),
        outSignal2=dict(),
    )
    outputs = JistIntensityMp2rageMasking.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import MRIsCALabel


def test_MRIsCALabel_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        aseg=dict(argstr='-aseg %s', ),
        canonsurf=dict(
            argstr='%s',
            mandatory=True,
            position=-3,
        ),
        classifier=dict(
            argstr='%s',
            mandatory=True,
            position=-2,
        ),
        copy_inputs=dict(),
        curv=dict(mandatory=True, ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        hemisphere=dict(
            argstr='%s',
            mandatory=True,
            position=-4,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        label=dict(argstr='-l %s', ),
        num_threads=dict(),
        out_file=dict(
            argstr='%s',
            hash_files=False,
            keep_extension=True,
            name_source=['hemisphere'],
            name_template='%s.aparc.annot',
            position=-1,
        ),
        seed=dict(argstr='-seed %d', ),
        smoothwm=dict(mandatory=True, ),
        subject_id=dict(
            argstr='%s',
            mandatory=True,
            position=-5,
            usedefault=True,
        ),
        subjects_dir=dict(),
        sulc=dict(mandatory=True, ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = MRIsCALabel.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MRIsCALabel_outputs():
    output_map = dict(out_file=dict(), )
    outputs = MRIsCALabel.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..dti import PicoPDFs


def test_PicoPDFs_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        directmap=dict(argstr='-directmap', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr='< %s',
            mandatory=True,
            position=1,
        ),
        inputmodel=dict(
            argstr='-inputmodel %s',
            position=2,
            usedefault=True,
        ),
        luts=dict(
            argstr='-luts %s',
            mandatory=True,
        ),
        maxcomponents=dict(
            argstr='-maxcomponents %d',
            units='NA',
        ),
        numpds=dict(
            argstr='-numpds %d',
            units='NA',
        ),
        out_file=dict(
            argstr='> %s',
            genfile=True,
            position=-1,
        ),
        pdf=dict(
            argstr='-pdf %s',
            position=4,
            usedefault=True,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = PicoPDFs.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_PicoPDFs_outputs():
    output_map = dict(pdfs=dict(), )
    outputs = PicoPDFs.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..segmentation import LaplacianThickness


def test_LaplacianThickness_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        dT=dict(
            argstr='dT=%d',
            position=6,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        input_gm=dict(
            argstr='%s',
            copyfile=True,
            mandatory=True,
            position=2,
        ),
        input_wm=dict(
            argstr='%s',
            copyfile=True,
            mandatory=True,
            position=1,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        opt_tolerance=dict(
            argstr='optional-laplacian-tolerance=%d',
            position=8,
        ),
        output_image=dict(
            argstr='%s',
            genfile=True,
            hash_files=False,
            position=3,
        ),
        prior_thickness=dict(
            argstr='priorthickval=%d',
            position=5,
        ),
        smooth_param=dict(
            argstr='smoothparam=%d',
            position=4,
        ),
        sulcus_prior=dict(
            argstr='use-sulcus-prior',
            position=7,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = LaplacianThickness.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_LaplacianThickness_outputs():
    output_map = dict(output_image=dict(), )
    outputs = LaplacianThickness.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from ....pipeline import engine as pe
from ....interfaces import fsl as fsl
from ....interfaces import freesurfer as fs
from ....interfaces import meshfix as mf
from ....interfaces import io as nio
from ....interfaces import utility as niu
from ....algorithms import misc as misc
from ....interfaces.utility import Function
from ....workflows.misc.utils import region_list_from_volume, id_list_from_lookup_table
import os


def get_aparc_aseg(files):
    """Return the aparc+aseg.mgz file"""
    for name in files:
        if 'aparc+aseg' in name:
            return name
    raise ValueError('aparc+aseg.mgz not found')


def create_getmask_flow(name='getmask', dilate_mask=True):
    """Registers a source file to freesurfer space and create a brain mask in
    source space

    Requires fsl tools for initializing registration

    Parameters
    ----------

    name : string
        name of workflow
    dilate_mask : boolean
        indicates whether to dilate mask or not

    Example
    -------

    >>> getmask = create_getmask_flow()
    >>> getmask.inputs.inputspec.source_file = 'mean.nii'
    >>> getmask.inputs.inputspec.subject_id = 's1'
    >>> getmask.inputs.inputspec.subjects_dir = '.'
    >>> getmask.inputs.inputspec.contrast_type = 't2'


    Inputs::

           inputspec.source_file : reference image for mask generation
           inputspec.subject_id : freesurfer subject id
           inputspec.subjects_dir : freesurfer subjects directory
           inputspec.contrast_type : MR contrast of reference image

    Outputs::

           outputspec.mask_file : binary mask file in reference image space
           outputspec.reg_file : registration file that maps reference image to
                                 freesurfer space
           outputspec.reg_cost : cost of registration (useful for detecting misalignment)
    """
    """
    Initialize the workflow
    """

    getmask = pe.Workflow(name=name)
    """
    Define the inputs to the workflow.
    """

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'source_file', 'subject_id', 'subjects_dir', 'contrast_type'
        ]),
        name='inputspec')
    """
    Define all the nodes of the workflow:

    fssource: used to retrieve aseg.mgz
    threshold : binarize aseg
    register : coregister source file to freesurfer space
    voltransform: convert binarized aseg to source file space
    """

    fssource = pe.Node(nio.FreeSurferSource(), name='fssource')
    threshold = pe.Node(fs.Binarize(min=0.5, out_type='nii'), name='threshold')
    register = pe.MapNode(
        fs.BBRegister(init='fsl'), iterfield=['source_file'], name='register')
    voltransform = pe.MapNode(
        fs.ApplyVolTransform(inverse=True),
        iterfield=['source_file', 'reg_file'],
        name='transform')
    """
    Connect the nodes
    """

    getmask.connect([(inputnode, fssource, [
        ('subject_id', 'subject_id'), ('subjects_dir', 'subjects_dir')
    ]), (inputnode, register,
         [('source_file', 'source_file'), ('subject_id', 'subject_id'),
          ('subjects_dir', 'subjects_dir'),
          ('contrast_type', 'contrast_type')]), (inputnode, voltransform, [
              ('subjects_dir', 'subjects_dir'), ('source_file', 'source_file')
          ]), (fssource, threshold, [(('aparc_aseg', get_aparc_aseg),
                                      'in_file')]),
                     (register, voltransform, [('out_reg_file', 'reg_file')]),
                     (threshold, voltransform, [('binary_file',
                                                 'target_file')])])
    """
    Add remaining nodes and connections

    dilate : dilate the transformed file in source space
    threshold2 : binarize transformed file
    """

    threshold2 = pe.MapNode(
        fs.Binarize(min=0.5, out_type='nii'),
        iterfield=['in_file'],
        name='threshold2')
    if dilate_mask:
        threshold2.inputs.dilate = 1
    getmask.connect([(voltransform, threshold2, [('transformed_file',
                                                  'in_file')])])
    """
    Setup an outputnode that defines relevant inputs of the workflow.
    """

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["mask_file", "reg_file", "reg_cost"]),
        name="outputspec")
    getmask.connect([
        (register, outputnode, [("out_reg_file", "reg_file")]),
        (register, outputnode, [("min_cost_file", "reg_cost")]),
        (threshold2, outputnode, [("binary_file", "mask_file")]),
    ])
    return getmask


def create_get_stats_flow(name='getstats', withreg=False):
    """Retrieves stats from labels

    Parameters
    ----------

    name : string
        name of workflow
    withreg : boolean
        indicates whether to register source to label

    Example
    -------


    Inputs::

           inputspec.source_file : reference image for mask generation
           inputspec.label_file : label file from which to get ROIs

           (optionally with registration)
           inputspec.reg_file : bbreg file (assumes reg from source to label
           inputspec.inverse : boolean whether to invert the registration
           inputspec.subjects_dir : freesurfer subjects directory

    Outputs::

           outputspec.stats_file : stats file
    """
    """
    Initialize the workflow
    """

    getstats = pe.Workflow(name=name)
    """
    Define the inputs to the workflow.
    """

    if withreg:
        inputnode = pe.Node(
            niu.IdentityInterface(fields=[
                'source_file', 'label_file', 'reg_file', 'subjects_dir'
            ]),
            name='inputspec')
    else:
        inputnode = pe.Node(
            niu.IdentityInterface(fields=['source_file', 'label_file']),
            name='inputspec')

    statnode = pe.MapNode(
        fs.SegStats(),
        iterfield=['segmentation_file', 'in_file'],
        name='segstats')
    """
    Convert between source and label spaces if registration info is provided

    """
    if withreg:
        voltransform = pe.MapNode(
            fs.ApplyVolTransform(inverse=True),
            iterfield=['source_file', 'reg_file'],
            name='transform')
        getstats.connect(inputnode, 'reg_file', voltransform, 'reg_file')
        getstats.connect(inputnode, 'source_file', voltransform, 'source_file')
        getstats.connect(inputnode, 'label_file', voltransform, 'target_file')
        getstats.connect(inputnode, 'subjects_dir', voltransform,
                         'subjects_dir')

        def switch_labels(inverse, transform_output, source_file, label_file):
            if inverse:
                return transform_output, source_file
            else:
                return label_file, transform_output

        chooser = pe.MapNode(
            niu.Function(
                input_names=[
                    'inverse', 'transform_output', 'source_file', 'label_file'
                ],
                output_names=['label_file', 'source_file'],
                function=switch_labels),
            iterfield=['transform_output', 'source_file'],
            name='chooser')
        getstats.connect(inputnode, 'source_file', chooser, 'source_file')
        getstats.connect(inputnode, 'label_file', chooser, 'label_file')
        getstats.connect(inputnode, 'inverse', chooser, 'inverse')
        getstats.connect(voltransform, 'transformed_file', chooser,
                         'transform_output')
        getstats.connect(chooser, 'label_file', statnode, 'segmentation_file')
        getstats.connect(chooser, 'source_file', statnode, 'in_file')
    else:
        getstats.connect(inputnode, 'label_file', statnode,
                         'segmentation_file')
        getstats.connect(inputnode, 'source_file', statnode, 'in_file')
    """
    Setup an outputnode that defines relevant inputs of the workflow.
    """

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["stats_file"]), name="outputspec")
    getstats.connect([
        (statnode, outputnode, [("summary_file", "stats_file")]),
    ])
    return getstats


def create_tessellation_flow(name='tessellate', out_format='stl'):
    """Tessellates the input subject's aseg.mgz volume and returns
    the surfaces for each region in stereolithic (.stl) format

    Example
    -------
    >>> from nipype.workflows.smri.freesurfer import create_tessellation_flow
    >>> tessflow = create_tessellation_flow()
    >>> tessflow.inputs.inputspec.subject_id = 'subj1'
    >>> tessflow.inputs.inputspec.subjects_dir = '.'
    >>> tessflow.inputs.inputspec.lookup_file = 'FreeSurferColorLUT.txt' # doctest: +SKIP
    >>> tessflow.run()  # doctest: +SKIP


    Inputs::

           inputspec.subject_id : freesurfer subject id
           inputspec.subjects_dir : freesurfer subjects directory
           inputspec.lookup_file : lookup file from freesurfer directory

    Outputs::

           outputspec.meshes : output region meshes in (by default) stereolithographic (.stl) format
    """
    """
    Initialize the workflow
    """

    tessflow = pe.Workflow(name=name)
    """
    Define the inputs to the workflow.
    """

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['subject_id', 'subjects_dir', 'lookup_file']),
        name='inputspec')
    """
    Define all the nodes of the workflow:

      fssource: used to retrieve aseg.mgz
      mri_convert : converts aseg.mgz to aseg.nii
      tessellate : tessellates regions in aseg.mgz
      surfconvert : converts regions to stereolithographic (.stl) format
      smoother: smooths the tessellated regions

    """

    fssource = pe.Node(nio.FreeSurferSource(), name='fssource')
    volconvert = pe.Node(fs.MRIConvert(out_type='nii'), name='volconvert')
    tessellate = pe.MapNode(
        fs.MRIMarchingCubes(),
        iterfield=['label_value', 'out_file'],
        name='tessellate')
    surfconvert = pe.MapNode(
        fs.MRIsConvert(out_datatype='stl'),
        iterfield=['in_file'],
        name='surfconvert')
    smoother = pe.MapNode(
        mf.MeshFix(), iterfield=['in_file1'], name='smoother')
    if out_format == 'gii':
        stl_to_gifti = pe.MapNode(
            fs.MRIsConvert(out_datatype=out_format),
            iterfield=['in_file'],
            name='stl_to_gifti')
    smoother.inputs.save_as_stl = True
    smoother.inputs.laplacian_smoothing_steps = 1

    region_list_from_volume_interface = Function(
        input_names=["in_file"],
        output_names=["region_list"],
        function=region_list_from_volume)

    id_list_from_lookup_table_interface = Function(
        input_names=["lookup_file", "region_list"],
        output_names=["id_list"],
        function=id_list_from_lookup_table)

    region_list_from_volume_node = pe.Node(
        interface=region_list_from_volume_interface,
        name='region_list_from_volume_node')
    id_list_from_lookup_table_node = pe.Node(
        interface=id_list_from_lookup_table_interface,
        name='id_list_from_lookup_table_node')
    """
    Connect the nodes
    """

    tessflow.connect([
        (inputnode, fssource, [('subject_id', 'subject_id'),
                               ('subjects_dir', 'subjects_dir')]),
        (fssource, volconvert, [('aseg', 'in_file')]),
        (volconvert, region_list_from_volume_node, [('out_file', 'in_file')]),
        (region_list_from_volume_node, tessellate, [('region_list',
                                                     'label_value')]),
        (region_list_from_volume_node, id_list_from_lookup_table_node,
         [('region_list', 'region_list')]),
        (inputnode, id_list_from_lookup_table_node, [('lookup_file',
                                                      'lookup_file')]),
        (id_list_from_lookup_table_node, tessellate, [('id_list',
                                                       'out_file')]),
        (fssource, tessellate, [('aseg', 'in_file')]),
        (tessellate, surfconvert, [('surface', 'in_file')]),
        (surfconvert, smoother, [('converted', 'in_file1')]),
    ])
    """
    Setup an outputnode that defines relevant inputs of the workflow.
    """

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["meshes"]), name="outputspec")

    if out_format == 'gii':
        tessflow.connect([
            (smoother, stl_to_gifti, [("mesh_file", "in_file")]),
        ])
        tessflow.connect([
            (stl_to_gifti, outputnode, [("converted", "meshes")]),
        ])
    else:
        tessflow.connect([
            (smoother, outputnode, [("mesh_file", "meshes")]),
        ])
    return tessflow


def copy_files(in_files, out_files):
    """
    Create a function to copy a file that can be modified by a following node
    without changing the original file
    """
    import shutil
    import sys
    if len(in_files) != len(out_files):
        print(
            "ERROR: Length of input files must be identical to the length of "
            + "outrput files to be copied")
        sys.exit(-1)
    for i, in_file in enumerate(in_files):
        out_file = out_files[i]
        print("copying {0} to {1}".format(in_file, out_file))
        shutil.copy(in_file, out_file)
    return out_files


def copy_file(in_file, out_file=None):
    """
    Create a function to copy a file that can be modified by a following node
    without changing the original file.
    """
    import os
    import shutil
    if out_file is None:
        out_file = os.path.join(os.getcwd(), os.path.basename(in_file))
    if type(in_file) is list and len(in_file) == 1:
        in_file = in_file[0]
    out_file = os.path.abspath(out_file)
    in_file = os.path.abspath(in_file)
    print("copying {0} to {1}".format(in_file, out_file))
    shutil.copy(in_file, out_file)
    return out_file


def mkdir_p(path):
    import errno
    import os
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def getdefaultconfig(exitonfail=False, rb_date="2014-08-21"):
    config = {
        'custom_atlas': None,
        'cw256': False,
        'field_strength': '1.5T',
        'fs_home': checkenv(exitonfail),
        'longitudinal': False,
        'long_base': None,
        'openmp': None,
        'plugin_args': None,
        'qcache': False,
        'queue': None,
        'recoding_file': None,
        'src_subject_id': 'fsaverage',
        'th3': True
    }

    config['src_subject_dir'] = os.path.join(config['fs_home'], 'subjects',
                                             config['src_subject_id'])
    config['awk_file'] = os.path.join(config['fs_home'], 'bin',
                                      'extract_talairach_avi_QA.awk')
    config['registration_template'] = os.path.join(
        config['fs_home'], 'average', 'RB_all_{0}.gca'.format(rb_date))
    config['registration_template_withskull'] = os.path.join(
        config['fs_home'], 'average',
        'RB_all_withskull_{0}.gca'.format(rb_date))
    for hemi in ('lh', 'rh'):
        config['{0}_atlas'.format(hemi)] = os.path.join(
            config['fs_home'], 'average',
            '{0}.average.curvature.filled.buckner40.tif'.format(hemi))
        config['{0}_classifier'.format(hemi)] = os.path.join(
            config['fs_home'], 'average',
            '{0}.curvature.buckner40.filled.desikan_killiany.2010-03-25.gcs'.
            format(hemi))
        config['{0}_classifier2'.format(hemi)] = os.path.join(
            config['fs_home'], 'average',
            '{0}.destrieux.simple.2009-07-29.gcs'.format(hemi))
        config['{0}_classifier3'.format(hemi)] = os.path.join(
            config['fs_home'], 'average', '{0}.DKTatlas40.gcs'.format(hemi))
    config['LookUpTable'] = os.path.join(config['fs_home'], 'ASegStatsLUT.txt')
    config['WMLookUpTable'] = os.path.join(config['fs_home'],
                                           'WMParcStatsLUT.txt')
    config['AvgColorTable'] = os.path.join(config['fs_home'], 'average',
                                           'colortable_BA.txt')

    return config


def checkenv(exitonfail=False):
    """Check for the necessary FS environment variables"""
    import sys
    fs_home = os.environ.get('FREESURFER_HOME')
    path = os.environ.get('PATH')
    print("FREESURFER_HOME: {0}".format(fs_home))
    if fs_home is None:
        msg = "please set FREESURFER_HOME before running the workflow"
    elif not os.path.isdir(fs_home):
        msg = "FREESURFER_HOME must be set to a valid directory before running this workflow"
    elif os.path.join(fs_home, 'bin') not in path.replace('//', '/'):
        print(path)
        msg = "Could not find necessary executable in path"
        setupscript = os.path.join(fs_home, 'SetUpFreeSurfer.sh')
        if os.path.isfile(setupscript):
            print("Please source the setup script before running the workflow:"
                  + "\nsource {0}".format(setupscript))
        else:
            print(
                "Please ensure that FREESURFER_HOME is set to a valid fs " +
                "directory and source the necessary SetUpFreeSurfer.sh script before running "
                + "this workflow")
    else:
        return fs_home

    if exitonfail:
        print("ERROR: " + msg)
        sys.exit(2)
    else:
        print("Warning: " + msg)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os.path as op

from ..base import (CommandLineInputSpec, CommandLine, traits, TraitedSpec,
                    File, isdefined, Undefined)
from .base import MRTrix3BaseInputSpec, MRTrix3Base


class ResponseSDInputSpec(MRTrix3BaseInputSpec):
    algorithm = traits.Enum(
        'msmt_5tt',
        'dhollander',
        'tournier',
        'tax',
        argstr='%s',
        position=1,
        mandatory=True,
        desc='response estimation algorithm (multi-tissue)')
    in_file = File(
        exists=True,
        argstr='%s',
        position=-5,
        mandatory=True,
        desc='input DWI image')
    mtt_file = File(argstr='%s', position=-4, desc='input 5tt image')
    wm_file = File(
        'wm.txt',
        argstr='%s',
        position=-3,
        usedefault=True,
        desc='output WM response text file')
    gm_file = File(
        argstr='%s', position=-2, desc='output GM response text file')
    csf_file = File(
        argstr='%s', position=-1, desc='output CSF response text file')
    in_mask = File(
        exists=True, argstr='-mask %s', desc='provide initial mask image')
    max_sh = traits.Int(
        8, usedefault=True,
        argstr='-lmax %d',
        desc='maximum harmonic degree of response function')


class ResponseSDOutputSpec(TraitedSpec):
    wm_file = File(argstr='%s', desc='output WM response text file')
    gm_file = File(argstr='%s', desc='output GM response text file')
    csf_file = File(argstr='%s', desc='output CSF response text file')


class ResponseSD(MRTrix3Base):
    """
    Estimate response function(s) for spherical deconvolution using the specified algorithm.

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> resp = mrt.ResponseSD()
    >>> resp.inputs.in_file = 'dwi.mif'
    >>> resp.inputs.algorithm = 'tournier'
    >>> resp.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> resp.cmdline                               # doctest: +ELLIPSIS
    'dwi2response tournier -fslgrad bvecs bvals -lmax 8 dwi.mif wm.txt'
    >>> resp.run()                                 # doctest: +SKIP
    """

    _cmd = 'dwi2response'
    input_spec = ResponseSDInputSpec
    output_spec = ResponseSDOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['wm_file'] = op.abspath(self.inputs.wm_file)
        if self.inputs.gm_file != Undefined:
            outputs['gm_file'] = op.abspath(self.inputs.gm_file)
        if self.inputs.csf_file != Undefined:
            outputs['csf_file'] = op.abspath(self.inputs.csf_file)
        return outputs


class ACTPrepareFSLInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=-2,
        desc='input anatomical image')

    out_file = File(
        'act_5tt.mif',
        argstr='%s',
        mandatory=True,
        position=-1,
        usedefault=True,
        desc='output file after processing')


class ACTPrepareFSLOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='the output response file')


class ACTPrepareFSL(CommandLine):
    """
    Generate anatomical information necessary for Anatomically
    Constrained Tractography (ACT).

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> prep = mrt.ACTPrepareFSL()
    >>> prep.inputs.in_file = 'T1.nii.gz'
    >>> prep.cmdline                               # doctest: +ELLIPSIS
    'act_anat_prepare_fsl T1.nii.gz act_5tt.mif'
    >>> prep.run()                                 # doctest: +SKIP
    """

    _cmd = 'act_anat_prepare_fsl'
    input_spec = ACTPrepareFSLInputSpec
    output_spec = ACTPrepareFSLOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = op.abspath(self.inputs.out_file)
        return outputs


class ReplaceFSwithFIRSTInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=-4,
        desc='input anatomical image')
    in_t1w = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=-3,
        desc='input T1 image')
    in_config = File(
        exists=True,
        argstr='%s',
        position=-2,
        desc='connectome configuration file')

    out_file = File(
        'aparc+first.mif',
        argstr='%s',
        mandatory=True,
        position=-1,
        usedefault=True,
        desc='output file after processing')


class ReplaceFSwithFIRSTOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='the output response file')


class ReplaceFSwithFIRST(CommandLine):
    """
    Replace deep gray matter structures segmented with FSL FIRST in a
    FreeSurfer parcellation.

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> prep = mrt.ReplaceFSwithFIRST()
    >>> prep.inputs.in_file = 'aparc+aseg.nii'
    >>> prep.inputs.in_t1w = 'T1.nii.gz'
    >>> prep.inputs.in_config = 'mrtrix3_labelconfig.txt'
    >>> prep.cmdline                               # doctest: +ELLIPSIS
    'fs_parc_replace_sgm_first aparc+aseg.nii T1.nii.gz \
mrtrix3_labelconfig.txt aparc+first.mif'
    >>> prep.run()                                 # doctest: +SKIP
    """

    _cmd = 'fs_parc_replace_sgm_first'
    input_spec = ReplaceFSwithFIRSTInputSpec
    output_spec = ReplaceFSwithFIRSTOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = op.abspath(self.inputs.out_file)
        return outputs



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .diffusion import create_camino_dti_pipeline
from .connectivity_mapping import create_connectivity_pipeline
from .group_connectivity import create_group_connectivity_pipeline



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import QwarpPlusMinus


def test_QwarpPlusMinus_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        base_file=dict(
            argstr='-base %s',
            copyfile=False,
            mandatory=True,
        ),
        blur=dict(argstr='-blur %s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        minpatch=dict(argstr='-minpatch %d', ),
        nopadWARP=dict(argstr='-nopadWARP', ),
        noweight=dict(argstr='-noweight', ),
        pblur=dict(argstr='-pblur %s', ),
        source_file=dict(
            argstr='-source %s',
            copyfile=False,
            mandatory=True,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = QwarpPlusMinus.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_QwarpPlusMinus_outputs():
    output_map = dict(
        base_warp=dict(),
        source_warp=dict(),
        warped_base=dict(),
        warped_source=dict(),
    )
    outputs = QwarpPlusMinus.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
import os
from ....interfaces.utility import Function, IdentityInterface
from ....pipeline import engine as pe  # pypeline engine
from ....interfaces.freesurfer import Label2Label, Label2Annot, ParcellationStats
from ....interfaces.io import DataGrabber
from ....interfaces.utility import Merge


def create_ba_maps_wf(name="Brodmann_Area_Maps",
                      th3=True,
                      exvivo=True,
                      entorhinal=True):
    # Brodmann Area Maps (BA Maps) and Hinds V1 Atlas
    inputs = [
        'lh_sphere_reg', 'rh_sphere_reg', 'lh_white', 'rh_white', 'lh_pial',
        'rh_pial', 'lh_orig', 'rh_orig', 'transform', 'lh_thickness',
        'rh_thickness', 'lh_cortex_label', 'rh_cortex_label', 'brainmask',
        'aseg', 'ribbon', 'wm', 'src_subject_id', 'src_subject_dir',
        'color_table'
    ]

    inputspec = pe.Node(IdentityInterface(fields=inputs), name="inputspec")

    ba_WF = pe.Workflow(name=name)

    ba_outputs = [
        'lh_BAMaps_stats', 'lh_color', 'lh_BAMaps_labels',
        'lh_BAMaps_annotation', 'lh_thresh_BAMaps_stats', 'lh_thresh_color',
        'lh_thresh_BAMaps_labels', 'lh_thresh_BAMaps_annotation',
        'rh_BAMaps_stats', 'rh_color', 'rh_BAMaps_labels',
        'rh_BAMaps_annotation', 'rh_thresh_BAMaps_stats', 'rh_thresh_color',
        'rh_thresh_BAMaps_labels', 'rh_thresh_BAMaps_annotation'
    ]

    outputspec = pe.Node(
        IdentityInterface(fields=ba_outputs), name="outputspec")

    labels = [
        "BA1", "BA2", "BA3a", "BA3b", "BA4a", "BA4p", "BA6", "BA44", "BA45",
        "V1", "V2", "MT", "perirhinal"
    ]
    if entorhinal:
        labels.insert(-1, 'entorhinal')
    for hemisphere in ['lh', 'rh']:
        for threshold in [True, False]:
            field_template = dict(
                sphere_reg='surf/{0}.sphere.reg'.format(hemisphere),
                white='surf/{0}.white'.format(hemisphere))

            out_files = list()
            source_fields = list()
            if threshold:
                for label in labels:
                    if label == 'perirhinal' and not entorhinal:
                        # versions < 6.0 do not use thresh.perirhinal
                        continue
                    if exvivo:
                        out_file = '{0}.{1}_exvivo.thresh.label'.format(
                            hemisphere, label)
                    else:
                        out_file = '{0}.{1}.thresh.label'.format(
                            hemisphere, label)
                    out_files.append(out_file)
                    field_template[label] = 'label/' + out_file
                    source_fields.append(label)
                node_name = 'BA_Maps_' + hemisphere + '_Thresh'
            else:
                for label in labels:
                    if exvivo:
                        out_file = '{0}.{1}_exvivo.label'.format(
                            hemisphere, label)
                    else:
                        out_file = '{0}.{1}.label'.format(hemisphere, label)

                    out_files.append(out_file)
                    field_template[label] = 'label/' + out_file
                    source_fields.append(label)
                node_name = 'BA_Maps_' + hemisphere

            source_subject = pe.Node(
                DataGrabber(outfields=source_fields + ['sphere_reg', 'white']),
                name=node_name + "_srcsubject")
            source_subject.inputs.template = '*'
            source_subject.inputs.sort_filelist = False
            source_subject.inputs.field_template = field_template
            ba_WF.connect([(inputspec, source_subject, [('src_subject_dir',
                                                         'base_directory')])])

            merge_labels = pe.Node(
                Merge(len(out_files)), name=node_name + "_Merge")
            for i, label in enumerate(source_fields):
                ba_WF.connect([(source_subject, merge_labels,
                                [(label, 'in{0}'.format(i + 1))])])

            node = pe.MapNode(
                Label2Label(),
                name=node_name + '_Label2Label',
                iterfield=['source_label', 'out_file'])
            node.inputs.hemisphere = hemisphere
            node.inputs.out_file = out_files
            node.inputs.copy_inputs = True

            ba_WF.connect(
                [(merge_labels, node, [('out', 'source_label')]),
                 (source_subject, node, [('sphere_reg', 'source_sphere_reg'),
                                         ('white', 'source_white')]),
                 (inputspec, node, [('src_subject_id', 'source_subject')])])

            label2annot = pe.Node(Label2Annot(), name=node_name + '_2_Annot')
            label2annot.inputs.hemisphere = hemisphere
            label2annot.inputs.verbose_off = True
            label2annot.inputs.keep_max = True
            label2annot.inputs.copy_inputs = True

            stats_node = pe.Node(
                ParcellationStats(), name=node_name + '_Stats')
            stats_node.inputs.hemisphere = hemisphere
            stats_node.inputs.mgz = True
            stats_node.inputs.th3 = th3
            stats_node.inputs.surface = 'white'
            stats_node.inputs.tabular_output = True
            stats_node.inputs.copy_inputs = True

            if threshold:
                label2annot.inputs.out_annot = "BA_exvivo.thresh"
                ba_WF.connect(
                    [(stats_node, outputspec,
                      [('out_color', '{0}_thresh_color'.format(hemisphere)),
                       ('out_table',
                        '{0}_thresh_BAMaps_stats'.format(hemisphere))]),
                     (label2annot, outputspec,
                      [('out_file',
                        '{0}_thresh_BAMaps_annotation'.format(hemisphere))]),
                     (node, outputspec,
                      [('out_file',
                        '{0}_thresh_BAMaps_labels'.format(hemisphere))])])
            else:
                label2annot.inputs.out_annot = "BA_exvivo"
                ba_WF.connect(
                    [(stats_node, outputspec,
                      [('out_color', '{0}_color'.format(hemisphere)),
                       ('out_table', '{0}_BAMaps_stats'.format(hemisphere))]),
                     (label2annot, outputspec,
                      [('out_file',
                        '{0}_BAMaps_annotation'.format(hemisphere))]),
                     (node, outputspec,
                      [('out_file', '{0}_BAMaps_labels'.format(hemisphere))])])

            ba_WF.connect(
                [(inputspec, node, [
                    ('{0}_sphere_reg'.format(hemisphere), 'sphere_reg'),
                    ('{0}_white'.format(hemisphere), 'white'),
                ]), (node, label2annot, [('out_file', 'in_labels')]),
                 (inputspec, label2annot,
                  [('{0}_orig'.format(hemisphere), 'orig'),
                   ('color_table', 'color_table')]), (label2annot, stats_node,
                                                      [('out_file',
                                                        'in_annotation')]),
                 (inputspec, stats_node,
                  [('{0}_thickness'.format(hemisphere),
                    'thickness'), ('{0}_cortex_label'.format(hemisphere),
                                   'cortex_label'), ('lh_white', 'lh_white'),
                   ('rh_white', 'rh_white'), ('lh_pial', 'lh_pial'),
                   ('rh_pial', 'rh_pial'), ('transform',
                                            'transform'), ('brainmask',
                                                           'brainmask'),
                   ('aseg', 'aseg'), ('wm', 'wm'), ('ribbon', 'ribbon')])])

    return ba_WF, ba_outputs



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from ....pipeline import engine as pe
from ....interfaces import mne as mne
from ....interfaces import freesurfer as fs
from ....interfaces import utility as niu


def create_bem_flow(name='bem', out_format='stl'):
    """Uses MNE's Watershed algorithm to create Boundary Element Meshes (BEM)
     for a subject's brain, inner/outer skull, and skin. The surfaces are
     returned in the desired (by default, stereolithic .stl) format.

    Example
    -------
    >>> from nipype.workflows.smri.freesurfer import create_bem_flow
    >>> bemflow = create_bem_flow()
    >>> bemflow.inputs.inputspec.subject_id = 'subj1'
    >>> bemflow.inputs.inputspec.subjects_dir = '.'
    >>> bemflow.run()  # doctest: +SKIP


    Inputs::

           inputspec.subject_id : freesurfer subject id
           inputspec.subjects_dir : freesurfer subjects directory

    Outputs::

           outputspec.meshes : output boundary element meshes in (by default)
                               stereolithographic (.stl) format
    """
    """
    Initialize the workflow
    """

    bemflow = pe.Workflow(name=name)
    """
    Define the inputs to the workflow.
    """

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['subject_id', 'subjects_dir']),
        name='inputspec')
    """
    Define all the nodes of the workflow:

      fssource: used to retrieve aseg.mgz
      mri_convert : converts aseg.mgz to aseg.nii
      tessellate : tessellates regions in aseg.mgz
      surfconvert : converts regions to stereolithographic (.stl) format

    """

    watershed_bem = pe.Node(interface=mne.WatershedBEM(), name='WatershedBEM')

    surfconvert = pe.MapNode(
        fs.MRIsConvert(out_datatype=out_format),
        iterfield=['in_file'],
        name='surfconvert')
    """
    Connect the nodes
    """

    bemflow.connect([
        (inputnode, watershed_bem, [('subject_id', 'subject_id'),
                                    ('subjects_dir', 'subjects_dir')]),
        (watershed_bem, surfconvert, [('mesh_files', 'in_file')]),
    ])
    """
    Setup an outputnode that defines relevant inputs of the workflow.
    """

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["meshes"]), name="outputspec")
    bemflow.connect([
        (surfconvert, outputnode, [("converted", "meshes")]),
    ])
    return bemflow



# -*- coding: utf-8 -*-
import os
from time import sleep

import nipype.interfaces.base as nib
import pytest
import nipype.pipeline.engine as pe

from nipype.pipeline.plugins.somaflow import soma_not_loaded


class InputSpec(nib.TraitedSpec):
    input1 = nib.traits.Int(desc='a random int')
    input2 = nib.traits.Int(desc='a random int')


class OutputSpec(nib.TraitedSpec):
    output1 = nib.traits.List(nib.traits.Int, desc='outputs')


class SomaTestInterface(nib.BaseInterface):
    input_spec = InputSpec
    output_spec = OutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['output1'] = [1, self.inputs.input1]
        return outputs


@pytest.mark.skipif(soma_not_loaded, reason="soma not loaded")
def test_run_somaflow(tmpdir):
    tmpdir.chdir()

    pipe = pe.Workflow(name='pipe')
    mod1 = pe.Node(interface=SomaTestInterface(), name='mod1')
    mod2 = pe.MapNode(
        interface=SomaTestInterface(), iterfield=['input1'], name='mod2')
    pipe.connect([(mod1, mod2, [('output1', 'input1')])])
    pipe.base_dir = os.getcwd()
    mod1.inputs.input1 = 1
    execgraph = pipe.run(plugin="SomaFlow")
    names = [
        '.'.join((node._hierarchy, node.name)) for node in execgraph.nodes()
    ]
    node = list(execgraph.nodes())[names.index('pipe.mod1')]
    result = node.get_output('output1')
    assert result == [1, 1]



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Nipype base interfaces
----------------------

This module defines the API of all nipype interfaces.

"""
from .core import (Interface, BaseInterface, SimpleInterface, CommandLine,
                   StdOutCommandLine, MpiCommandLine, SEMLikeCommandLine,
                   LibraryBaseInterface, PackageInfo)

from .specs import (BaseTraitedSpec, TraitedSpec, DynamicTraitedSpec,
                    BaseInterfaceInputSpec, CommandLineInputSpec,
                    StdOutCommandLineInputSpec)

from .traits_extension import (
    traits, Undefined, TraitDictObject, TraitListObject, TraitError, isdefined,
    File, Directory, Str, DictStrStr, has_metadata, ImageFile,
    OutputMultiObject, InputMultiObject,
    OutputMultiPath, InputMultiPath)

from .support import (Bunch, InterfaceResult, load_template,
                      NipypeInterfaceError)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import DWI2Tensor


def test_DWI2Tensor_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        debug=dict(
            argstr='-debug',
            position=1,
        ),
        encoding_file=dict(
            argstr='-grad %s',
            position=2,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        ignore_slice_by_volume=dict(
            argstr='-ignoreslices %s',
            position=2,
            sep=' ',
        ),
        ignore_volumes=dict(
            argstr='-ignorevolumes %s',
            position=2,
            sep=' ',
        ),
        in_file=dict(
            argstr='%s',
            mandatory=True,
            position=-2,
        ),
        out_filename=dict(
            argstr='%s',
            name_source='in_file',
            name_template='%s_tensor.mif',
            output_name='tensor',
            position=-1,
        ),
        quiet=dict(
            argstr='-quiet',
            position=1,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = DWI2Tensor.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_DWI2Tensor_outputs():
    output_map = dict(tensor=dict(), )
    outputs = DWI2Tensor.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# coding: utf-8
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import open, str

import warnings

from ....pipeline import engine as pe
from ....interfaces import utility as niu
from ....interfaces import fsl as fsl


def create_dmri_preprocessing(name='dMRI_preprocessing',
                              use_fieldmap=True,
                              fieldmap_registration=False):
    """
    Creates a workflow that chains the necessary pipelines to
    correct for motion, eddy currents, and, if selected, susceptibility
    artifacts in EPI dMRI sequences.

    .. deprecated:: 0.9.3
      Use :func:`nipype.workflows.dmri.preprocess.epi.all_fmb_pipeline` or
      :func:`nipype.workflows.dmri.preprocess.epi.all_peb_pipeline` instead.


    .. warning:: This workflow rotates the b-vectors, so please be
      advised that not all the dicom converters ensure the consistency between the resulting
      nifti orientation and the b matrix table (e.g. dcm2nii checks it).


    Example
    -------

    >>> nipype_dmri_preprocess = create_dmri_preprocessing('nipype_dmri_prep')
    >>> nipype_dmri_preprocess.inputs.inputnode.in_file = 'diffusion.nii'
    >>> nipype_dmri_preprocess.inputs.inputnode.in_bvec = 'diffusion.bvec'
    >>> nipype_dmri_preprocess.inputs.inputnode.ref_num = 0
    >>> nipype_dmri_preprocess.inputs.inputnode.fieldmap_mag = 'magnitude.nii'
    >>> nipype_dmri_preprocess.inputs.inputnode.fieldmap_pha = 'phase.nii'
    >>> nipype_dmri_preprocess.inputs.inputnode.te_diff = 2.46
    >>> nipype_dmri_preprocess.inputs.inputnode.epi_echospacing = 0.77
    >>> nipype_dmri_preprocess.inputs.inputnode.epi_rev_encoding = False
    >>> nipype_dmri_preprocess.inputs.inputnode.pi_accel_factor = True
    >>> nipype_dmri_preprocess.run() # doctest: +SKIP


    Inputs::

        inputnode.in_file - The diffusion data
        inputnode.in_bvec - The b-matrix file, in FSL format and consistent with the in_file orientation
        inputnode.ref_num - The reference volume (a b=0 volume in dMRI)
        inputnode.fieldmap_mag - The magnitude of the fieldmap
        inputnode.fieldmap_pha - The phase difference of the fieldmap
        inputnode.te_diff - TE increment used (in msec.) on the fieldmap acquisition (generally 2.46ms for 3T scanners)
        inputnode.epi_echospacing - The EPI EchoSpacing parameter (in msec.)
        inputnode.epi_rev_encoding - True if reverse encoding was used (generally False)
        inputnode.pi_accel_factor - Parallel imaging factor (aka GRAPPA acceleration factor)
        inputnode.vsm_sigma - Sigma (in mm.) of the gaussian kernel used for in-slice smoothing of the deformation field (voxel shift map, vsm)


    Outputs::

        outputnode.dmri_corrected
        outputnode.bvec_rotated


    Optional arguments::

        use_fieldmap - True if there are fieldmap files that should be used (default True)
        fieldmap_registration - True if registration to fieldmap should be performed (default False)


    """

    warnings.warn(
        ('This workflow is deprecated from v.1.0.0, use of available '
         'nipype.workflows.dmri.preprocess.epi.all_*'), DeprecationWarning)

    pipeline = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'in_file', 'in_bvec', 'ref_num', 'fieldmap_mag', 'fieldmap_pha',
            'te_diff', 'epi_echospacing', 'epi_rev_encoding',
            'pi_accel_factor', 'vsm_sigma'
        ]),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['dmri_corrected', 'bvec_rotated']),
        name='outputnode')

    motion = create_motion_correct_pipeline()
    eddy = create_eddy_correct_pipeline()

    if use_fieldmap:  # we have a fieldmap, so lets use it (yay!)
        susceptibility = create_epidewarp_pipeline(
            fieldmap_registration=fieldmap_registration)

        pipeline.connect(
            [(inputnode, motion, [('in_file', 'inputnode.in_file'),
                                  ('in_bvec', 'inputnode.in_bvec'),
                                  ('ref_num', 'inputnode.ref_num')]),
             (inputnode, eddy,
              [('ref_num', 'inputnode.ref_num')]), (motion, eddy, [
                  ('outputnode.motion_corrected', 'inputnode.in_file')
              ]), (eddy, susceptibility,
                   [('outputnode.eddy_corrected', 'inputnode.in_file')]),
             (inputnode, susceptibility,
              [('ref_num', 'inputnode.ref_num'), ('fieldmap_mag',
                                                  'inputnode.fieldmap_mag'),
               ('fieldmap_pha', 'inputnode.fieldmap_pha'),
               ('te_diff', 'inputnode.te_diff'), ('epi_echospacing',
                                                  'inputnode.epi_echospacing'),
               ('epi_rev_encoding',
                'inputnode.epi_rev_encoding'), ('pi_accel_factor',
                                                'inputnode.pi_accel_factor'),
               ('vsm_sigma', 'inputnode.vsm_sigma')]), (motion, outputnode, [
                   ('outputnode.out_bvec', 'bvec_rotated')
               ]), (susceptibility, outputnode, [('outputnode.epi_corrected',
                                                  'dmri_corrected')])])
    else:  # we don't have a fieldmap, so we just carry on without it :(
        pipeline.connect([(inputnode, motion, [
            ('in_file', 'inputnode.in_file'), ('in_bvec', 'inputnode.in_bvec'),
            ('ref_num', 'inputnode.ref_num')
        ]), (inputnode, eddy, [('ref_num', 'inputnode.ref_num')]),
                          (motion, eddy, [('outputnode.motion_corrected',
                                           'inputnode.in_file')]),
                          (motion, outputnode,
                           [('outputnode.out_bvec',
                             'bvec_rotated')]), (eddy, outputnode,
                                                 [('outputnode.eddy_corrected',
                                                   'dmri_corrected')])])

    return pipeline


def create_motion_correct_pipeline(name='motion_correct'):
    """Creates a pipeline that corrects for motion artifact in dMRI sequences.
    It takes a series of diffusion weighted images and rigidly co-registers
    them to one reference image. Finally, the b-matrix is rotated accordingly
    (Leemans et al. 2009 - http://www.ncbi.nlm.nih.gov/pubmed/19319973),
    making use of the rotation matrix obtained by FLIRT.


    .. deprecated:: 0.9.3
      Use :func:`nipype.workflows.dmri.preprocess.epi.hmc_pipeline` instead.


    .. warning:: This workflow rotates the b-vectors, so please be adviced
      that not all the dicom converters ensure the consistency between the resulting
      nifti orientation and the b matrix table (e.g. dcm2nii checks it).


    Example
    -------

    >>> nipype_motioncorrect = create_motion_correct_pipeline('nipype_motioncorrect')
    >>> nipype_motioncorrect.inputs.inputnode.in_file = 'diffusion.nii'
    >>> nipype_motioncorrect.inputs.inputnode.in_bvec = 'diffusion.bvec'
    >>> nipype_motioncorrect.inputs.inputnode.ref_num = 0
    >>> nipype_motioncorrect.run() # doctest: +SKIP

    Inputs::

        inputnode.in_file
        inputnode.ref_num
        inputnode.in_bvec

    Outputs::

        outputnode.motion_corrected
        outputnode.out_bvec

    """

    warnings.warn(
        ('This workflow is deprecated from v.1.0.0, use '
         'nipype.workflows.dmri.preprocess.epi.hmc_pipeline instead'),
        DeprecationWarning)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_file', 'ref_num', 'in_bvec']),
        name='inputnode')

    pipeline = pe.Workflow(name=name)

    split = pe.Node(fsl.Split(dimension='t'), name='split')
    pick_ref = pe.Node(niu.Select(), name='pick_ref')
    coregistration = pe.MapNode(
        fsl.FLIRT(no_search=True, interp='spline', padding_size=1, dof=6),
        name='coregistration',
        iterfield=['in_file'])
    rotate_bvecs = pe.Node(
        niu.Function(
            input_names=['in_bvec', 'in_matrix'],
            output_names=['out_file'],
            function=_rotate_bvecs),
        name='rotate_b_matrix')
    merge = pe.Node(fsl.Merge(dimension='t'), name='merge')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['motion_corrected', 'out_bvec']),
        name='outputnode')

    pipeline.connect(
        [(inputnode, split, [('in_file', 'in_file')]),
         (split, pick_ref, [('out_files', 'inlist')]), (inputnode, pick_ref, [
             ('ref_num', 'index')
         ]), (split, coregistration,
              [('out_files', 'in_file')]), (inputnode, rotate_bvecs,
                                            [('in_bvec', 'in_bvec')]),
         (coregistration, rotate_bvecs,
          [('out_matrix_file', 'in_matrix')]), (pick_ref, coregistration,
                                                [('out', 'reference')]),
         (coregistration, merge,
          [('out_file', 'in_files')]), (merge, outputnode, [
              ('merged_file', 'motion_corrected')
          ]), (rotate_bvecs, outputnode, [('out_file', 'out_bvec')])])

    return pipeline


def create_eddy_correct_pipeline(name='eddy_correct'):
    """

    .. deprecated:: 0.9.3
      Use :func:`nipype.workflows.dmri.preprocess.epi.ecc_pipeline` instead.


    Creates a pipeline that replaces eddy_correct script in FSL. It takes a
    series of diffusion weighted images and linearly co-registers them to one
    reference image. No rotation of the B-matrix is performed, so this pipeline
    should be executed after the motion correction pipeline.

    Example
    -------

    >>> nipype_eddycorrect = create_eddy_correct_pipeline('nipype_eddycorrect')
    >>> nipype_eddycorrect.inputs.inputnode.in_file = 'diffusion.nii'
    >>> nipype_eddycorrect.inputs.inputnode.ref_num = 0
    >>> nipype_eddycorrect.run() # doctest: +SKIP

    Inputs::

        inputnode.in_file
        inputnode.ref_num

    Outputs::

        outputnode.eddy_corrected
    """

    warnings.warn(
        ('This workflow is deprecated from v.1.0.0, use '
         'nipype.workflows.dmri.preprocess.epi.ecc_pipeline instead'),
        DeprecationWarning)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_file', 'ref_num']), name='inputnode')

    pipeline = pe.Workflow(name=name)

    split = pe.Node(fsl.Split(dimension='t'), name='split')
    pick_ref = pe.Node(niu.Select(), name='pick_ref')
    coregistration = pe.MapNode(
        fsl.FLIRT(no_search=True, padding_size=1, interp='trilinear'),
        name='coregistration',
        iterfield=['in_file'])
    merge = pe.Node(fsl.Merge(dimension='t'), name='merge')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['eddy_corrected']), name='outputnode')

    pipeline.connect([(inputnode, split, [('in_file', 'in_file')]),
                      (split, pick_ref,
                       [('out_files', 'inlist')]), (inputnode, pick_ref,
                                                    [('ref_num', 'index')]),
                      (split, coregistration,
                       [('out_files', 'in_file')]), (pick_ref, coregistration,
                                                     [('out', 'reference')]),
                      (coregistration, merge,
                       [('out_file', 'in_files')]), (merge, outputnode,
                                                     [('merged_file',
                                                       'eddy_corrected')])])
    return pipeline


def fieldmap_correction(name='fieldmap_correction', nocheck=False):
    """

    .. deprecated:: 0.9.3
      Use :func:`nipype.workflows.dmri.preprocess.epi.sdc_fmb` instead.


    Fieldmap-based retrospective correction of EPI images for the susceptibility distortion
    artifact (Jezzard et al., 1995). Fieldmap images are assumed to be already registered
    to EPI data, and a brain mask is required.

    Replaces the former workflow, still available as create_epidewarp_pipeline().  The difference
    with respect the epidewarp pipeline is that now the workflow uses the new fsl_prepare_fieldmap
    available as of FSL 5.0.


    Example
    -------

    >>> nipype_epicorrect = fieldmap_correction('nipype_epidewarp')
    >>> nipype_epicorrect.inputs.inputnode.in_file = 'diffusion.nii'
    >>> nipype_epicorrect.inputs.inputnode.in_mask = 'brainmask.nii'
    >>> nipype_epicorrect.inputs.inputnode.fieldmap_pha = 'phase.nii'
    >>> nipype_epicorrect.inputs.inputnode.fieldmap_mag = 'magnitude.nii'
    >>> nipype_epicorrect.inputs.inputnode.te_diff = 2.46
    >>> nipype_epicorrect.inputs.inputnode.epi_echospacing = 0.77
    >>> nipype_epicorrect.inputs.inputnode.encoding_direction = 'y'
    >>> nipype_epicorrect.run() # doctest: +SKIP

    Inputs::

        inputnode.in_file - The volume acquired with EPI sequence
        inputnode.in_mask - A brain mask
        inputnode.fieldmap_pha - The phase difference map from the fieldmapping, registered to in_file
        inputnode.fieldmap_mag - The magnitud maps (usually 4D, one magnitude per GRE scan)
                                 from the fieldmapping, registered to in_file
        inputnode.te_diff - Time difference in msec. between TE in ms of the fieldmapping (usually a GRE sequence).
        inputnode.epi_echospacing - The effective echo spacing (aka dwell time) in msec. of the EPI sequence. If
                                    EPI was acquired with parallel imaging, then the effective echo spacing is
                                    eff_es = es / acc_factor.
        inputnode.encoding_direction - The phase encoding direction in EPI acquisition (default y)
        inputnode.vsm_sigma - Sigma value of the gaussian smoothing filter applied to the vsm (voxel shift map)


    Outputs::

        outputnode.epi_corrected
        outputnode.out_vsm

    """

    warnings.warn(('This workflow is deprecated from v.1.0.0, use '
                   'nipype.workflows.dmri.preprocess.epi.sdc_fmb instead'),
                  DeprecationWarning)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'in_file', 'in_mask', 'fieldmap_pha', 'fieldmap_mag', 'te_diff',
            'epi_echospacing', 'vsm_sigma', 'encoding_direction'
        ]),
        name='inputnode')

    pipeline = pe.Workflow(name=name)

    # Keep first frame from magnitude
    select_mag = pe.Node(
        fsl.utils.ExtractROI(t_size=1, t_min=0), name='select_magnitude')

    # Mask magnitude (it is required by PreparedFieldMap)
    mask_mag = pe.Node(fsl.maths.ApplyMask(), name='mask_magnitude')

    # Run fsl_prepare_fieldmap
    fslprep = pe.Node(fsl.PrepareFieldmap(), name='prepare_fieldmap')

    if nocheck:
        fslprep.inputs.nocheck = True

    # Use FUGUE to generate the voxel shift map (vsm)
    vsm = pe.Node(fsl.FUGUE(save_shift=True), name='generate_vsm')

    # VSM demean is not anymore present in the epi_reg script
    # vsm_mean = pe.Node(niu.Function(input_names=['in_file', 'mask_file', 'in_unwarped'], output_names=[
    #                   'out_file'], function=_vsm_remove_mean), name='vsm_mean_shift')

    # fugue_epi
    dwi_split = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_files'],
            function=_split_dwi),
        name='dwi_split')

    # 'fugue -i %s -u %s --loadshift=%s --mask=%s' % ( vol_name, out_vol_name, vsm_name, mask_name )
    dwi_applyxfm = pe.MapNode(
        fsl.FUGUE(icorr=True, save_shift=False),
        iterfield=['in_file'],
        name='dwi_fugue')
    # Merge back all volumes
    dwi_merge = pe.Node(fsl.utils.Merge(dimension='t'), name='dwi_merge')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['epi_corrected', 'out_vsm']),
        name='outputnode')

    pipeline.connect(
        [(inputnode, select_mag,
          [('fieldmap_mag', 'in_file')]), (inputnode, fslprep, [
              ('fieldmap_pha', 'in_phase'), ('te_diff', 'delta_TE')
          ]), (inputnode, mask_mag,
               [('in_mask', 'mask_file')]), (select_mag, mask_mag,
                                             [('roi_file', 'in_file')]),
         (mask_mag, fslprep, [('out_file', 'in_magnitude')]), (fslprep, vsm, [
             ('out_fieldmap', 'phasemap_in_file')
         ]), (inputnode,
              vsm, [('fieldmap_mag',
                     'in_file'), ('encoding_direction', 'unwarp_direction'),
                    (('te_diff', _ms2sec), 'asym_se_time'),
                    ('vsm_sigma', 'smooth2d'), (('epi_echospacing', _ms2sec),
                                                'dwell_time')]),
         (mask_mag, vsm, [('out_file', 'mask_file')]), (inputnode, dwi_split, [
             ('in_file', 'in_file')
         ]), (dwi_split, dwi_applyxfm,
              [('out_files', 'in_file')]), (mask_mag, dwi_applyxfm,
                                            [('out_file', 'mask_file')]),
         (vsm, dwi_applyxfm,
          [('shift_out_file', 'shift_in_file')]), (inputnode, dwi_applyxfm, [
              ('encoding_direction', 'unwarp_direction')
          ]), (dwi_applyxfm, dwi_merge,
               [('unwarped_file', 'in_files')]), (dwi_merge, outputnode, [
                   ('merged_file', 'epi_corrected')
               ]), (vsm, outputnode, [('shift_out_file', 'out_vsm')])])

    return pipeline


def topup_correction(name='topup_correction'):
    """

    .. deprecated:: 0.9.3
      Use :func:`nipype.workflows.dmri.preprocess.epi.sdc_peb` instead.


    Corrects for susceptibilty distortion of EPI images when one reverse encoding dataset has
    been acquired


    Example
    -------

    >>> nipype_epicorrect = topup_correction('nipype_topup')
    >>> nipype_epicorrect.inputs.inputnode.in_file_dir = 'epi.nii'
    >>> nipype_epicorrect.inputs.inputnode.in_file_rev = 'epi_rev.nii'
    >>> nipype_epicorrect.inputs.inputnode.encoding_direction = ['y', 'y-']
    >>> nipype_epicorrect.inputs.inputnode.ref_num = 0
    >>> nipype_epicorrect.run() # doctest: +SKIP


    Inputs::

        inputnode.in_file_dir - EPI volume acquired in 'forward' phase encoding
        inputnode.in_file_rev - EPI volume acquired in 'reversed' phase encoding
        inputnode.encoding_direction - Direction encoding of in_file_dir
        inputnode.ref_num - Identifier of the reference volumes (usually B0 volume)


    Outputs::

        outputnode.epi_corrected


    """

    warnings.warn(('This workflow is deprecated from v.1.0.0, use '
                   'nipype.workflows.dmri.preprocess.epi.sdc_peb instead'),
                  DeprecationWarning)

    pipeline = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'in_file_dir', 'in_file_rev', 'encoding_direction',
            'readout_times', 'ref_num'
        ]),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'out_fieldcoef', 'out_movpar', 'out_enc_file', 'epi_corrected'
        ]),
        name='outputnode')

    b0_dir = pe.Node(fsl.ExtractROI(t_size=1), name='b0_1')
    b0_rev = pe.Node(fsl.ExtractROI(t_size=1), name='b0_2')
    combin = pe.Node(niu.Merge(2), name='merge')
    combin2 = pe.Node(niu.Merge(2), name='merge2')
    merged = pe.Node(fsl.Merge(dimension='t'), name='b0_comb')

    topup = pe.Node(fsl.TOPUP(), name='topup')
    applytopup = pe.Node(fsl.ApplyTOPUP(in_index=[1, 2]), name='applytopup')

    pipeline.connect(
        [(inputnode, b0_dir, [('in_file_dir', 'in_file'), ('ref_num',
                                                           't_min')]),
         (inputnode, b0_rev,
          [('in_file_rev',
            'in_file'), ('ref_num', 't_min')]), (inputnode, combin2, [
                ('in_file_dir', 'in1'), ('in_file_rev', 'in2')
            ]), (b0_dir, combin, [('roi_file', 'in1')]), (b0_rev, combin, [
                ('roi_file', 'in2')
            ]), (combin, merged, [('out', 'in_files')]),
         (merged, topup, [('merged_file', 'in_file')]), (inputnode, topup, [
             ('encoding_direction', 'encoding_direction'), ('readout_times',
                                                            'readout_times')
         ]), (topup, applytopup, [('out_fieldcoef', 'in_topup_fieldcoef'),
                                  ('out_movpar', 'in_topup_movpar'),
                                  ('out_enc_file', 'encoding_file')]),
         (combin2, applytopup, [('out', 'in_files')]), (topup, outputnode, [
             ('out_fieldcoef', 'out_fieldcoef'), ('out_movpar', 'out_movpar'),
             ('out_enc_file', 'out_enc_file')
         ]), (applytopup, outputnode, [('out_corrected', 'epi_corrected')])])

    return pipeline


def create_epidewarp_pipeline(name='epidewarp', fieldmap_registration=False):
    """
    Replaces the epidewarp.fsl script (http://www.nmr.mgh.harvard.edu/~greve/fbirn/b0/epidewarp.fsl)
    for susceptibility distortion correction of dMRI & fMRI acquired with EPI sequences and the fieldmap
    information (Jezzard et al., 1995) using FSL's FUGUE. The registration to the (warped) fieldmap
    (strictly following the original script) is available using fieldmap_registration=True.


    .. warning:: This workflow makes use of ``epidewarp.fsl`` a script of FSL deprecated long
      time ago. The use of this workflow is not recommended, use
      :func:`nipype.workflows.dmri.preprocess.epi.sdc_fmb` instead.


    Example
    -------

    >>> nipype_epicorrect = create_epidewarp_pipeline('nipype_epidewarp', fieldmap_registration=False)
    >>> nipype_epicorrect.inputs.inputnode.in_file = 'diffusion.nii'
    >>> nipype_epicorrect.inputs.inputnode.fieldmap_mag = 'magnitude.nii'
    >>> nipype_epicorrect.inputs.inputnode.fieldmap_pha = 'phase.nii'
    >>> nipype_epicorrect.inputs.inputnode.te_diff = 2.46
    >>> nipype_epicorrect.inputs.inputnode.epi_echospacing = 0.77
    >>> nipype_epicorrect.inputs.inputnode.epi_rev_encoding = False
    >>> nipype_epicorrect.inputs.inputnode.ref_num = 0
    >>> nipype_epicorrect.inputs.inputnode.pi_accel_factor = 1.0
    >>> nipype_epicorrect.run() # doctest: +SKIP

    Inputs::

        inputnode.in_file - The volume acquired with EPI sequence
        inputnode.fieldmap_mag - The magnitude of the fieldmap
        inputnode.fieldmap_pha - The phase difference of the fieldmap
        inputnode.te_diff - Time difference between TE in ms.
        inputnode.epi_echospacing - The echo spacing (aka dwell time) in the EPI sequence
        inputnode.epi_ph_encoding_dir - The phase encoding direction in EPI acquisition (default y)
        inputnode.epi_rev_encoding - True if it is acquired with reverse encoding
        inputnode.pi_accel_factor - Acceleration factor used for EPI parallel imaging (GRAPPA)
        inputnode.vsm_sigma - Sigma value of the gaussian smoothing filter applied to the vsm (voxel shift map)
        inputnode.ref_num - The reference volume (B=0 in dMRI or a central frame in fMRI)


    Outputs::

        outputnode.epi_corrected


    Optional arguments::

        fieldmap_registration - True if registration to fieldmap should be done (default False)

    """

    warnings.warn(('This workflow reproduces a deprecated FSL script.'),
                  DeprecationWarning)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'in_file', 'fieldmap_mag', 'fieldmap_pha', 'te_diff',
            'epi_echospacing', 'epi_ph_encoding_dir', 'epi_rev_encoding',
            'pi_accel_factor', 'vsm_sigma', 'ref_num', 'unwarp_direction'
        ]),
        name='inputnode')

    pipeline = pe.Workflow(name=name)

    # Keep first frame from magnitude
    select_mag = pe.Node(
        fsl.utils.ExtractROI(t_size=1, t_min=0), name='select_magnitude')

    # mask_brain
    mask_mag = pe.Node(fsl.BET(mask=True), name='mask_magnitude')
    mask_mag_dil = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_file'],
            function=_dilate_mask),
        name='mask_dilate')

    # Compute dwell time
    dwell_time = pe.Node(
        niu.Function(
            input_names=['dwell_time', 'pi_factor', 'is_reverse_encoding'],
            output_names=['dwell_time'],
            function=_compute_dwelltime),
        name='dwell_time')

    # Normalize phase diff to be [-pi, pi)
    norm_pha = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_file'],
            function=_prepare_phasediff),
        name='normalize_phasediff')
    # Execute FSL PRELUDE: prelude -p %s -a %s -o %s -f -v -m %s
    prelude = pe.Node(fsl.PRELUDE(process3d=True), name='phase_unwrap')
    fill_phase = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_file'],
            function=_fill_phase),
        name='fill_phasediff')

    # to assure that vsm is same dimension as mag. The input only affects the output dimension.
    # The content of the input has no effect on the vsm. The de-warped mag volume is
    # meaningless and will be thrown away
    # fugue -i %s -u %s -p %s --dwell=%s --asym=%s --mask=%s --saveshift=%s %
    # ( mag_name, magdw_name, ph_name, esp, tediff, mask_name, vsmmag_name)
    vsm = pe.Node(fsl.FUGUE(save_shift=True), name='generate_vsm')
    vsm_mean = pe.Node(
        niu.Function(
            input_names=['in_file', 'mask_file', 'in_unwarped'],
            output_names=['out_file'],
            function=_vsm_remove_mean),
        name='vsm_mean_shift')

    # fugue_epi
    dwi_split = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_files'],
            function=_split_dwi),
        name='dwi_split')
    # 'fugue -i %s -u %s --loadshift=%s --mask=%s' % ( vol_name, out_vol_name, vsm_name, mask_name )
    dwi_applyxfm = pe.MapNode(
        fsl.FUGUE(icorr=True, save_shift=False),
        iterfield=['in_file'],
        name='dwi_fugue')
    # Merge back all volumes
    dwi_merge = pe.Node(fsl.utils.Merge(dimension='t'), name='dwi_merge')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['epi_corrected']), name='outputnode')

    pipeline.connect(
        [(inputnode, dwell_time,
          [('epi_echospacing', 'dwell_time'), ('pi_accel_factor', 'pi_factor'),
           ('epi_rev_encoding',
            'is_reverse_encoding')]), (inputnode, select_mag, [('fieldmap_mag',
                                                                'in_file')]),
         (inputnode, norm_pha, [('fieldmap_pha',
                                 'in_file')]), (select_mag, mask_mag,
                                                [('roi_file', 'in_file')]),
         (mask_mag, mask_mag_dil,
          [('mask_file', 'in_file')]), (select_mag, prelude, [
              ('roi_file', 'magnitude_file')
          ]), (norm_pha, prelude,
               [('out_file', 'phase_file')]), (mask_mag_dil, prelude, [
                   ('out_file', 'mask_file')
               ]), (prelude, fill_phase,
                    [('unwrapped_phase_file', 'in_file')]), (inputnode, vsm, [
                        ('fieldmap_mag', 'in_file')
                    ]), (fill_phase, vsm, [('out_file', 'phasemap_in_file')]),
         (inputnode, vsm, [(('te_diff', _ms2sec), 'asym_se_time'),
                           ('vsm_sigma', 'smooth2d')]), (dwell_time, vsm, [
                               (('dwell_time', _ms2sec), 'dwell_time')
                           ]), (mask_mag_dil, vsm, [('out_file',
                                                     'mask_file')]),
         (mask_mag_dil, vsm_mean,
          [('out_file', 'mask_file')]), (vsm, vsm_mean, [
              ('unwarped_file', 'in_unwarped'), ('shift_out_file', 'in_file')
          ]), (inputnode, dwi_split,
               [('in_file', 'in_file')]), (dwi_split, dwi_applyxfm, [
                   ('out_files', 'in_file')
               ]), (dwi_applyxfm, dwi_merge,
                    [('unwarped_file', 'in_files')]), (dwi_merge, outputnode,
                                                       [('merged_file',
                                                         'epi_corrected')])])

    if fieldmap_registration:
        """ Register magfw to example epi. There are some parameters here that may need to be tweaked. Should probably strip the mag
            Pre-condition: forward warp the mag in order to reg with func. What does mask do here?
        """
        # Select reference volume from EPI (B0 in dMRI and a middle frame in
        # fMRI)
        select_epi = pe.Node(fsl.utils.ExtractROI(t_size=1), name='select_epi')

        # fugue -i %s -w %s --loadshift=%s --mask=%s % ( mag_name, magfw_name,
        # vsmmag_name, mask_name ), log ) # Forward Map
        vsm_fwd = pe.Node(fsl.FUGUE(forward_warping=True), name='vsm_fwd')
        vsm_reg = pe.Node(
            fsl.FLIRT(
                bins=256,
                cost='corratio',
                dof=6,
                interp='spline',
                searchr_x=[-10, 10],
                searchr_y=[-10, 10],
                searchr_z=[-10, 10]),
            name='vsm_registration')
        # 'flirt -in %s -ref %s -out %s -init %s -applyxfm' % ( vsmmag_name, ref_epi, vsmmag_name, magfw_mat_out )
        vsm_applyxfm = pe.Node(
            fsl.ApplyXfm(interp='spline'), name='vsm_apply_xfm')
        # 'flirt -in %s -ref %s -out %s -init %s -applyxfm' % ( mask_name, ref_epi, mask_name, magfw_mat_out )
        msk_applyxfm = pe.Node(
            fsl.ApplyXfm(interp='nearestneighbour'), name='msk_apply_xfm')

        pipeline.connect(
            [(inputnode, select_epi,
              [('in_file', 'in_file'),
               ('ref_num', 't_min')]), (select_epi, vsm_reg, [('roi_file',
                                                               'reference')]),
             (vsm, vsm_fwd, [('shift_out_file', 'shift_in_file')]),
             (mask_mag_dil, vsm_fwd,
              [('out_file', 'mask_file')]), (inputnode, vsm_fwd, [
                  ('fieldmap_mag', 'in_file')
              ]), (vsm_fwd, vsm_reg,
                   [('warped_file', 'in_file')]), (vsm_reg, msk_applyxfm, [
                       ('out_matrix_file', 'in_matrix_file')
                   ]), (select_epi, msk_applyxfm, [('roi_file', 'reference')]),
             (mask_mag_dil, msk_applyxfm,
              [('out_file', 'in_file')]), (vsm_reg, vsm_applyxfm, [
                  ('out_matrix_file', 'in_matrix_file')
              ]), (select_epi, vsm_applyxfm,
                   [('roi_file', 'reference')]), (vsm_mean, vsm_applyxfm,
                                                  [('out_file', 'in_file')]),
             (msk_applyxfm, dwi_applyxfm,
              [('out_file', 'mask_file')]), (vsm_applyxfm, dwi_applyxfm,
                                             [('out_file', 'shift_in_file')])])
    else:
        pipeline.connect(
            [(mask_mag_dil, dwi_applyxfm, [('out_file', 'mask_file')]),
             (vsm_mean, dwi_applyxfm, [('out_file', 'shift_in_file')])])

    return pipeline


def _rotate_bvecs(in_bvec, in_matrix):
    import os
    import numpy as np

    name, fext = os.path.splitext(os.path.basename(in_bvec))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('./%s_rotated.bvec' % name)
    bvecs = np.loadtxt(in_bvec)
    new_bvecs = np.zeros(
        shape=bvecs.T.shape)  # pre-initialise array, 3 col format

    for i, vol_matrix in enumerate(in_matrix[0::]):  # start index at 0
        bvec = np.matrix(bvecs[:, i])
        rot = np.matrix(np.loadtxt(vol_matrix)[0:3, 0:3])
        new_bvecs[i] = (np.array(
            rot * bvec.T).T)[0]  # fill each volume with x,y,z as we go along
    np.savetxt(out_file, np.array(new_bvecs).T, fmt=b'%0.15f')
    return out_file


def _cat_logs(in_files):
    import shutil
    import os

    name, fext = os.path.splitext(os.path.basename(in_files[0]))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('./%s_ecclog.log' % name)
    with open(out_file, 'wb') as totallog:
        for i, fname in enumerate(in_files):
            totallog.write('\n\npreprocessing %d\n' % i)
            with open(fname) as inlog:
                for line in inlog:
                    totallog.write(line)
    return out_file


def _compute_dwelltime(dwell_time=0.68,
                       pi_factor=1.0,
                       is_reverse_encoding=False):
    dwell_time *= (1.0 / pi_factor)

    if is_reverse_encoding:
        dwell_time *= -1.0

    return dwell_time


def _effective_echospacing(dwell_time, pi_factor=1.0):
    dwelltime = 1.0e-3 * dwell_time * (1.0 / pi_factor)
    return dwelltime


def _prepare_phasediff(in_file):
    import nibabel as nb
    import os
    import numpy as np
    from nipype.utils import NUMPY_MMAP
    img = nb.load(in_file, mmap=NUMPY_MMAP)
    max_diff = np.max(img.get_data().reshape(-1))
    min_diff = np.min(img.get_data().reshape(-1))
    A = (2.0 * np.pi) / (max_diff - min_diff)
    B = np.pi - (A * max_diff)
    diff_norm = img.get_data() * A + B

    name, fext = os.path.splitext(os.path.basename(in_file))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('./%s_2pi.nii.gz' % name)
    nb.save(nb.Nifti1Image(diff_norm, img.affine, img.header), out_file)
    return out_file


def _dilate_mask(in_file, iterations=4):
    import nibabel as nb
    import scipy.ndimage as ndimage
    import os
    from nipype.utils import NUMPY_MMAP
    img = nb.load(in_file, mmap=NUMPY_MMAP)
    dilated_img = img.__class__(
        ndimage.binary_dilation(img.get_data(), iterations=iterations),
        img.affine, img.header)

    name, fext = os.path.splitext(os.path.basename(in_file))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('./%s_dil.nii.gz' % name)
    nb.save(dilated_img, out_file)
    return out_file


def _fill_phase(in_file):
    import nibabel as nb
    import os
    import numpy as np
    from nipype.utils import NUMPY_MMAP
    img = nb.load(in_file, mmap=NUMPY_MMAP)
    dumb_img = nb.Nifti1Image(np.zeros(img.shape), img.affine, img.header)
    out_nii = nb.funcs.concat_images((img, dumb_img))
    name, fext = os.path.splitext(os.path.basename(in_file))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('./%s_fill.nii.gz' % name)
    nb.save(out_nii, out_file)
    return out_file


def _vsm_remove_mean(in_file, mask_file, in_unwarped):
    import nibabel as nb
    import os
    import numpy as np
    import numpy.ma as ma
    from nipype.utils import NUMPY_MMAP
    img = nb.load(in_file, mmap=NUMPY_MMAP)
    msk = nb.load(mask_file, mmap=NUMPY_MMAP).get_data()
    img_data = img.get_data()
    img_data[msk == 0] = 0
    vsmmag_masked = ma.masked_values(img_data.reshape(-1), 0.0)
    vsmmag_masked = vsmmag_masked - vsmmag_masked.mean()
    masked_img = img.__class__(
        vsmmag_masked.reshape(img.shape), img.affine, img.header)
    name, fext = os.path.splitext(os.path.basename(in_file))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('./%s_demeaned.nii.gz' % name)
    nb.save(masked_img, out_file)
    return out_file


def _ms2sec(val):
    return val * 1e-3


def _split_dwi(in_file):
    import nibabel as nb
    import os
    from nipype.utils import NUMPY_MMAP
    out_files = []
    frames = nb.funcs.four_to_three(nb.load(in_file, mmap=NUMPY_MMAP))
    name, fext = os.path.splitext(os.path.basename(in_file))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    for i, frame in enumerate(frames):
        out_file = os.path.abspath('./%s_%03d.nii.gz' % (name, i))
        nb.save(frame, out_file)
        out_files.append(out_file)
    return out_files
